"""This module contains the main class of lshell, it is the class that is
responsible for the command line interface. It inherits from the cmd.Cmd class
from the Python standard library. It offers the default methods to add
security checks, logging, etc.
"""

import cmd
import sys
import os
import re
import signal
import readline

# import lshell specifics
from lshell.checkconfig import CheckConfig
from lshell import utils
from lshell import builtincmd
from lshell import messages
from lshell import sec
from lshell import completion
from lshell import variables
from lshell import policy as policy_mode
from lshell import audit


class ShellCmd(cmd.Cmd, object):
    """Main lshell CLI class"""

    def __init__(
        self,
        userconf,
        args,
        stdin=None,
        stdout=None,
        stderr=None,
        g_cmd=None,
        g_line=None,
    ):
        if stdin is None:
            self.stdin = sys.stdin
        else:
            self.stdin = stdin
        if stdout is None:
            self.stdout = sys.stdout
        else:
            self.stdout = stdout
        if stderr is None:
            self.stderr = sys.stderr
        else:
            self.stderr = stderr

        self.conf = userconf
        self.log = self.conf["logpath"]
        self.kill_jobs_at_exit = False

        # Set timer
        if self.conf["timer"] > 0:
            self.mytimer(self.conf["timer"])
        self.identchars = self.identchars + "+./-"
        self.log.error("Logged in")
        cmd.Cmd.__init__(self)

        # set prompt
        self.conf["promptprint"] = utils.updateprompt(os.getcwd(), self.conf)

        self.intro = self.conf["intro"]

        # initialize oldpwd variable to home directory
        self.conf["oldpwd"] = self.conf["home_path"]

        # initialize cli variables
        self.g_cmd = g_cmd
        self.g_line = g_line
        self.args = args

        # initialize return code
        self.retcode = 0

        # run overssh, if needed
        self.run_overssh()

    def __getattr__(self, attr):
        """This method actually takes care of all the called method that are
        not resolved (i.e not existing methods). It actually will simulate
        the existence of any method    entered in the 'allowed' variable list.

        e.g. You just have to add 'uname' in list of allowed commands in
        the 'allowed' variable, and lshell will react as if you had
        added a do_uname in the ShellCmd class!
        """

        # in case the configuration file has been modified, reload it
        if self.conf["config_mtime"] != os.path.getmtime(self.conf["configfile"]):
            session_id = self.conf.get("session_id")
            self.conf = CheckConfig(
                ["--config", self.conf["configfile"]], refresh=1
            ).returnconf()
            if session_id:
                self.conf["session_id"] = session_id
            self.conf["promptprint"] = utils.updateprompt(os.getcwd(), self.conf)
            self.log = self.conf["logpath"]

        if self.conf["timer"] > 0:
            self.mytimer(0)

        # replace $? with the exit code
        self.g_line = utils.replace_exit_code(self.g_line, self.retcode)

        if isinstance(self.conf["aliases"], dict):
            self.g_line = utils.get_aliases(self.g_line, self.conf["aliases"])

        self.log.info(f'CMD: "{self.g_line}"')

        self.retcode = utils.cmd_parse_execute(self.g_line, shell_context=self)

        self.g_cmd, self.g_arg, self.g_line = ["", "", ""]

        if self.conf["timer"] > 0:
            self.mytimer(self.conf["timer"])
        return object.__getattribute__(self, attr)

    def run_overssh(self):
        """This method checks if the user is trying to SCP a file onto the
        server. If this is the case, it checks if the user is allowed to use
        SCP or not, and    acts as requested. : )
        """
        def _execute_trusted_ssh_protocol(trusted_protocol=False):
            # Protocol commands are still validated in run_overssh, then
            # executed through the regular command path so policy settings
            # (path/sudo/allowed_cmd_path/env/umask) stay consistent.
            return utils.cmd_parse_execute(
                self.conf["ssh"],
                shell_context=self,
                trusted_protocol=trusted_protocol,
            )

        def _validate_ssh_command(check_path=True):
            ret_check_secure, self.conf = sec.check_secure(
                self.conf["ssh"], self.conf, strict=1, ssh=1
            )
            if ret_check_secure:
                self.ssh_warn("char/command over SSH", self.conf["ssh"])

            if check_path:
                ret_check_path, self.conf = sec.check_path(
                    self.conf["ssh"], self.conf, strict=1, ssh=1
                )
                if ret_check_path == 1:
                    self.ssh_warn("path over SSH", self.conf["ssh"])

        def _with_protocol_in_overssh(protocol_commands):
            overssh = list(self.conf.get("overssh", []))
            changed = False
            for item in protocol_commands:
                if item and item not in overssh:
                    overssh.append(item)
                    changed = True
            if changed:
                self.conf["overssh"] = overssh

        def _aliases_for_ssh_command():
            aliases = self.conf["aliases"]
            if self.conf.get("_auto_ls_alias") and isinstance(aliases, dict):
                aliases = dict(aliases)
                aliases.pop("ls", None)
            return aliases

        if "ssh" in self.conf:
            if "SSH_CLIENT" in os.environ and "SSH_TTY" not in os.environ:
                # Apply aliases consistently for all SSH command paths.
                self.conf["ssh"] = utils.get_aliases(
                    self.conf["ssh"], _aliases_for_ssh_command()
                ).strip()

                # check if sftp is requested and allowed
                if "sftp-server" in self.conf["ssh"]:
                    if self.conf["sftp"] == 1:
                        _with_protocol_in_overssh(
                            variables.TRUSTED_SFTP_PROTOCOL_BINARIES
                        )
                        # sftp-server binary path may live outside restricted
                        # user paths; keep command-level checks but skip path ACL.
                        _validate_ssh_command(check_path=False)
                        self.log.error("SFTP connect")
                        retcode = _execute_trusted_ssh_protocol(trusted_protocol=True)
                        self.log.error("SFTP disconnect")
                        sys.exit(retcode)
                    else:
                        self.log.error("*** forbidden SFTP connection")
                        audit.log_command_event(
                            self.conf,
                            self.conf["ssh"],
                            allowed=False,
                            reason="forbidden SFTP connection",
                        )
                        sys.exit(1)

                # check if scp is requested and allowed
                if self.conf["ssh"].startswith("scp "):
                    if self.conf["scp"] == 1 or "scp" in self.conf["overssh"]:
                        _with_protocol_in_overssh(["scp"])

                        if " -f " in self.conf["ssh"]:
                            # case scp download is allowed
                            if self.conf["scp_download"]:
                                self.log.error(f'SCP: GET "{self.conf["ssh"]}"')
                            # case scp download is forbidden
                            else:
                                self.log.error(
                                    f'SCP: download forbidden: "{self.conf["ssh"]}"'
                                )
                                audit.log_command_event(
                                    self.conf,
                                    self.conf["ssh"],
                                    allowed=False,
                                    reason="forbidden SCP download",
                                )
                                sys.exit(1)
                        elif " -t " in self.conf["ssh"]:
                            # case scp upload is allowed
                            if self.conf["scp_upload"]:
                                if "scpforce" in self.conf:
                                    cmdsplit = self.conf["ssh"].split(" ")
                                    scppath = os.path.realpath(cmdsplit[-1])
                                    forcedpath = os.path.realpath(self.conf["scpforce"])
                                    if scppath != forcedpath:
                                        self.log.error(
                                            f"SCP: forced SCP directory: {scppath}"
                                        )
                                        cmdsplit.pop(-1)
                                        cmdsplit.append(forcedpath)
                                        self.conf["ssh"] = " ".join(cmdsplit)
                                self.log.error(f'SCP: PUT "{self.conf["ssh"]}"')
                            # case scp upload is forbidden
                            else:
                                self.log.error(
                                    f'SCP: upload forbidden: "{self.conf["ssh"]}"'
                                )
                                audit.log_command_event(
                                    self.conf,
                                    self.conf["ssh"],
                                    allowed=False,
                                    reason="forbidden SCP upload",
                                )
                                sys.exit(1)
                        _validate_ssh_command()
                        retcode = _execute_trusted_ssh_protocol(trusted_protocol=False)
                        self.log.error("SCP disconnect")
                        sys.exit(retcode)
                    else:
                        self.ssh_warn("SCP connection", self.conf["ssh"], "scp")

                # check if command is in allowed overssh commands
                elif self.conf["ssh"]:
                    _validate_ssh_command()
                    self.log.error(f'Over SSH: "{self.conf["ssh"]}"')
                    # if command is "help"
                    if self.conf["ssh"] == "help":
                        self.do_help(None)
                        retcode = 0
                    else:
                        retcode = utils.cmd_parse_execute(
                            self.conf["ssh"], shell_context=self
                        )
                    self.log.error("Exited")
                    sys.exit(retcode)

                # else warn and log
                else:
                    self.ssh_warn("command over SSH", self.conf["ssh"])
            else:
                # case of local shell escapes (e.g. pager/editor invoking
                # the login shell with -c). Validate against normal policy.
                self.conf["ssh"] = utils.get_aliases(
                    self.conf["ssh"], _aliases_for_ssh_command()
                )
                ret_check_secure, self.conf = sec.check_secure(
                    self.conf["ssh"],
                    self.conf,
                    strict=self.conf["strict"],
                )
                if ret_check_secure:
                    self.log.error(f'*** forbidden shell escape: "{self.conf["ssh"]}"')
                    audit.log_command_event(
                        self.conf,
                        self.conf["ssh"],
                        allowed=False,
                        reason=audit.pop_decision_reason(
                            self.conf, "forbidden shell escape"
                        ),
                    )
                    sys.exit(1)

                self.log.error(f'Shell escape: "{self.conf["ssh"]}"')
                retcode = utils.cmd_parse_execute(
                    self.conf["ssh"], shell_context=self
                )
                self.log.error("Exited")
                sys.exit(retcode)
            return retcode

    def ssh_warn(self, message, command="", key=""):
        """log and warn if forbidden action over SSH"""
        audit.log_command_event(
            self.conf,
            command,
            allowed=False,
            reason=f"forbidden over SSH: {message}",
        )
        if key == "scp":
            self.log.critical(
                messages.get_message(self.conf, "forbidden_scp_over_ssh", message=message)
            )
            self.log.error(f"lshell: SCP command: {command}")
        else:
            self.log.critical(
                messages.get_message(
                    self.conf,
                    "forbidden_command_over_ssh",
                    message=message,
                    command=command,
                )
            )
        sys.stderr.write(messages.get_message(self.conf, "incident_reported") + "\n")
        self.log.error("Exited")
        sys.exit(1)

    def run_script_mode(self, script):
        """Process commands from a script."""
        with open(script, "r") as script_file:
            for line in script_file:
                line = line.strip()
                if line:
                    line = self.precmd(line)
                    stop = self.onecmd(line)
                    stop = self.postcmd(stop, line)
                    if stop:
                        sys.exit(1)

    def cmdloop(self, intro=None):
        """Repeatedly issue a prompt, accept input, parse an initial prefix
        off the received input, and dispatch to action methods, passing them
        the remainder of the line as argument.
        """

        if self.conf.get("script"):
            self.run_script_mode(self.conf["script"])
            return

        self.preloop()
        if self.use_rawinput and self.completekey:
            try:
                readline.read_history_file(self.conf["history_file"])
            except IOError:
                # if history file does not exist
                try:
                    open(self.conf["history_file"], "w").close()
                    readline.read_history_file(self.conf["history_file"])
                except IOError:
                    pass
            readline.set_history_length(self.conf["history_size"])
            readline.set_completer_delims(
                readline.get_completer_delims().replace("-", "")
            )
            self.old_completer = readline.get_completer()
            readline.set_completer(self.complete)
            readline.parse_and_bind(self.completekey + ": complete")
        try:
            if self.intro and isinstance(self.intro, str):
                self.stdout.write(f"{self.intro}\n")
            if self.conf["login_script"]:
                utils.cmd_parse_execute(self.conf["login_script"], shell_context=self)
            self.prompt2 = "> "  # PS2 prompt
            # for long commands, a user may escape the new line
            # by giving a bash like '\' character at the end of
            # the line. cmdloop() needs to recognize that and
            # create an appended line before sending it to onecmd()
            partial_line = ""
            stop = None
            while not stop:
                try:
                    # Check background jobs after each command
                    builtincmd.check_background_jobs()
                    if self.cmdqueue:
                        line = self.cmdqueue.pop(0)
                    else:
                        if self.use_rawinput:
                            try:
                                line = input(self.conf["promptprint"])
                            except EOFError:
                                line = "EOF"
                            except KeyboardInterrupt:
                                self.stdout.write("\n")
                                if partial_line:
                                    partial_line = ""
                                    self.conf["promptprint"] = utils.updateprompt(
                                        os.getcwd(), self.conf
                                    )
                                continue
                        else:
                            self.stdout.write(self.conf["promptprint"])
                            self.stdout.flush()
                            line = self.stdin.readline()
                            if not line:
                                line = "EOF"
                            else:
                                # chop \n
                                line = line[:-1]
                        if len(line) > 1 and line.startswith("\\"):
                            # implying previous partial line
                            line = line[:1].replace("\\", "", 1)
                        if partial_line:
                            line = partial_line + line
                        if line.endswith("\\"):
                            # continuation character. First partial line.
                            # We shall expect the command to continue in
                            # a new line. Change to bash like PS2 prompt to
                            # indicate this continuation to the user
                            partial_line = line.strip("\\")
                            self.conf["promptprint"] = self.prompt2  # switching to PS2
                            continue
                        elif line.count('"') % 2 != 0 or line.count("'") % 2 != 0:
                            # unclosed quotes detected
                            partial_line = line
                            self.conf["promptprint"] = self.prompt2  # switching to PS2
                            continue
                        partial_line = ""
                        self.conf["promptprint"] = utils.updateprompt(
                            os.getcwd(), self.conf
                        )
                    line = self.precmd(line)
                    stop = self.onecmd(line)
                    stop = self.postcmd(stop, line)
                except KeyboardInterrupt:
                    # Keep prompt handling local to cmdloop even when Ctrl+C
                    # races outside input() (e.g. background-job checks).
                    self.stdout.write("\n")
                    self.stdout.flush()
                    partial_line = ""
                    self.conf["promptprint"] = utils.updateprompt(
                        os.getcwd(), self.conf
                    )
                    continue
            self.postloop()
        finally:
            if self.use_rawinput and self.completekey:
                try:
                    readline.set_completer_delims(
                        readline.get_completer_delims().replace("-", "")
                    )
                    readline.set_completer(self.old_completer)
                except ImportError:
                    pass
            try:
                readline.write_history_file(self.conf["history_file"])
            except IOError:
                self.log.error(
                    f"WARN: couldn't write history to file {self.conf['history_file']}\n"
                )

    def complete(self, text, state):
        """Return the next possible completion for 'text'.
        If a command has not been entered, then complete against command list.
        Otherwise try to call complete_<command> to get list of completions.
        """
        if state == 0:
            origline = readline.get_line_buffer()
            line = origline.lstrip()
            # in case '|', ';', '&' used, take last part of line to complete
            line = re.split(r"&|\||;", line)[-1].lstrip()
            stripped = len(origline) - len(line)
            begidx = readline.get_begidx() - stripped
            endidx = readline.get_endidx() - stripped
            # complete with sudo allowed commands
            command = line.split(" ")[0]
            if command == "sudo" and len(line.split(" ")) <= 2:
                compfunc = completion.complete_sudo
            # complete with directories
            elif command == "cd":
                compfunc = completion.complete_change_dir
            # complete local relative commands from allowed entries like ./foo
            elif command.startswith("./"):
                compfunc = completion.completenames
            # complete with files and directories
            elif (
                len(line.split(" ")) > 1 and line.split(" ")[0] in self.conf["allowed"]
            ):
                compfunc = completion.complete_list_dir
            elif begidx > 0:
                cmd, args, _ = self.parseline(line)
                if cmd == "":
                    compfunc = completion.completedefault
                else:
                    try:
                        compfunc = getattr(self, "complete_" + cmd)
                    except AttributeError:
                        compfunc = (
                            completion.completenames
                            if cmd.startswith("./")
                            else completion.completedefault
                        )
                    # exception called when using './' completion
                    except IndexError:
                        compfunc = completion.completenames
                    # exception called when using './' completion
                    except TypeError:
                        compfunc = completion.completenames
            else:
                # call the lshell allowed commands completion
                compfunc = completion.completenames

            self.completion_matches = compfunc(self.conf, text, line, begidx, endidx)
        try:
            return self.completion_matches[state]
        except IndexError:
            return None

    def default(self, line):
        """This method overrides the original default method.
        It was originally used to warn when an unknown command was entered
        (e.g. *** Unknown syntax: blabla).
        It has been implemented in the __getattr__ method.
        So it has no use here. Its output is now empty.
        """
        self.stdout.write("")

    def completenames(self, text, line, *ignored):
        """Override the original completenames method."""
        return completion.completenames(self.conf, text, line, *ignored)

    def onecmd(self, line):
        """This method overrides the original onecmd method, to put the cmd,
        arg and line variables in class global variables: self.g_cmd,
        self.g_arg and self.g_line.
        Those variables are then used by the __getattr__ method
        """
        cmd, arg, line = self.parseline(line)
        self.g_cmd, self.g_arg, self.g_line = [cmd, arg, line]
        if not line:
            return self.emptyline()
        if cmd is None:
            return self.default(line)
        self.lastcmd = line
        if cmd == "":
            return self.default(line)
        else:
            try:
                func = getattr(self, "do_" + cmd)
            except AttributeError:
                return self.default(line)
            return func(arg)

    def emptyline(self):
        """This method overrides the original emptyline method, so it doesn't
        repeat the last command if last command was empty.
        I just found this annoying..
        """
        if self.lastcmd:
            return 0

    def do_help(self, arg=None):
        """This method overrides the original do_help method.
        Instead of printing out the that are documented or not, it returns the
        list of allowed commands when '?' or 'help' is entered.
        Of course, it doesn't override the help function: any help_* method
        will be called (e.g. help_help(self) )
        """

        # Get list of allowed commands, remove duplicate 'help' then sort
        list_tmp = list(dict.fromkeys(self.completenames("", "")).keys())
        list_tmp.sort()
        self.columnize(list_tmp)

    def do_EOF(self, arg=None):  # pylint: disable=invalid-name
        """Handle Ctrl+D / EOF exactly like exit."""
        return self.do_exit(arg)

    def do_quit(self, arg=None):
        """Handle quit exactly like exit."""
        return self.do_exit(arg)

    def do_policy_show(self, arg=None):
        """Show resolved policy values and optional decision for a command."""
        command_line = (arg or "").strip() or None
        username = self.conf.get("username")
        groups = policy_mode._resolve_user_groups(username, [])
        try:
            result = policy_mode.resolve_policy(
                self.conf["configfile"],
                username,
                groups,
            )
        except ValueError as exception:
            self.stderr.write(f"lshell: {exception}\n")
            return 1

        decision = None
        if command_line:
            decision = policy_mode.policy_command_decision(command_line, result["policy"])

        policy_mode.print_user_view(result, command_line, decision)
        if decision is None:
            return 0
        return 0 if decision["allowed"] else 2

    def do_exit(self, arg=None):
        """This method overrides the original do_exit method."""
        # Check for background jobs
        if hasattr(builtincmd, "BACKGROUND_JOBS") and builtincmd.BACKGROUND_JOBS:
            # Filter out completed jobs
            active_jobs = []
            for job_id, job in enumerate(builtincmd.BACKGROUND_JOBS, start=1):
                if job.poll() is None:
                    active_jobs.append((job_id, job))

            if active_jobs and self.kill_jobs_at_exit:
                for job_id, job in active_jobs:
                    try:
                        os.killpg(os.getpgid(job.pid), signal.SIGKILL)
                        builtincmd.BACKGROUND_JOBS.pop(job_id - 1)
                    except Exception as exception:
                        print(f"Failed to stop job [{job.pid}]: {exception}")
            else:
                # Warn the user and list the stopped jobs
                print(
                    "There are stopped jobs. Use 'jobs' to list them or 'exit' "
                    "to stop them and exit shell."
                )
                self.kill_jobs_at_exit = True
                return  # Return to the shell prompt instead of exiting

        # Proceed with exit if no active jobs or after stopping them
        if self.g_cmd == "EOF":
            self.stdout.write("\n")

        if self.conf["disable_exit"] != 1:
            sys.exit(0)

    def mytimer(self, timeout):
        """This function is kicks you out the the lshell after
        the 'timer' variable expires. 'timer' is set in seconds.
        """
        # set timer
        signal.signal(signal.SIGALRM, self._timererror)
        signal.alarm(timeout)

    def _timererror(self, signum, frame):
        raise LshellTimeOut("lshell timer timeout")


class LshellTimeOut(Exception):
    """Custom exception used for timer timeout"""

    def __init__(self, value="Timed Out"):
        self.value = value

    def __str__(self):
        return repr(self.value)



"""Functional tests for structured security audit logging."""

import json
import os
import tempfile
import unittest
from getpass import getuser

import pexpect


TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
CONFIG = f"{TOPDIR}/test/testfiles/test.conf"
LSHELL = f"{TOPDIR}/bin/lshell"
USER = getuser()
PROMPT = f"{USER}:~\\$"


class TestAuditFunctional(unittest.TestCase):
    """Validate ECS audit events from a real interactive shell session."""

    def _load_command_events(self, logfile):
        events = []
        with open(logfile, "r", encoding="utf-8") as handle:
            for line in handle:
                line = line.strip()
                if not line:
                    continue
                try:
                    payload = json.loads(line)
                except json.JSONDecodeError:
                    continue
                if payload.get("event.action") == "command_authorization":
                    events.append(payload)
        return events

    def test_security_audit_json_emits_allowed_and_denied_events(self):
        """Emit structured events with success/failure outcomes and reasons."""
        with tempfile.TemporaryDirectory(prefix="lshell-audit-log-") as log_dir:
            child = pexpect.spawn(
                f"{LSHELL} --config {CONFIG} --log {log_dir} --security_audit_json=1 "
                "--loglevel 4 --strict 0",
                encoding="utf-8",
                timeout=10,
            )
            try:
                child.expect(PROMPT)

                child.sendline("echo AUDIT_OK")
                child.expect(PROMPT)

                child.sendline("id")
                child.expect(PROMPT)

                child.sendline("exit")
                child.expect(pexpect.EOF)
            finally:
                child.close()

            logfile = os.path.join(log_dir, f"{USER}.log")
            self.assertTrue(os.path.exists(logfile))

            events = self._load_command_events(logfile)

            self.assertGreaterEqual(len(events), 2)

            allowed = [
                event
                for event in events
                if event.get("process.command_line") == "echo AUDIT_OK"
                and event.get("event.outcome") == "success"
            ]
            denied = [
                event
                for event in events
                if event.get("process.command_line") == "id"
                and event.get("event.outcome") == "failure"
            ]

            self.assertTrue(allowed, msg=f"missing allowed audit event in {events}")
            self.assertTrue(denied, msg=f"missing denied audit event in {events}")
            self.assertIn("unknown syntax", denied[0].get("event.reason", ""))

    def test_runtime_limit_denial_reason_is_machine_readable_in_audit(self):
        """Denied runtime-limit actions should emit machine-readable reason strings."""
        with tempfile.TemporaryDirectory(prefix="lshell-audit-log-") as log_dir:
            with tempfile.TemporaryDirectory(prefix="lshell-audit-session-") as session_dir:
                env = os.environ.copy()
                env["LSHELL_SESSION_DIR"] = session_dir
                child = pexpect.spawn(
                    f"{LSHELL} --config {CONFIG} --log {log_dir} --security_audit_json=1 "
                    "--loglevel 4 --strict 1 --forbidden \"[]\" --allowed \"['sleep']\" "
                    "--max_background_jobs 1",
                    encoding="utf-8",
                    timeout=10,
                    env=env,
                )
                try:
                    child.expect(PROMPT)
                    child.sendline("sleep 60 &")
                    child.expect(PROMPT)
                    child.sendline("sleep 60 &")
                    child.expect(PROMPT)
                    child.sendline("exit")
                    child.expect(PROMPT)
                    child.sendline("exit")
                    child.expect(pexpect.EOF)
                finally:
                    child.close()

            logfile = os.path.join(log_dir, f"{USER}.log")
            self.assertTrue(os.path.exists(logfile))
            events = self._load_command_events(logfile)

            denied = [
                event
                for event in events
                if event.get("event.outcome") == "failure"
                and "runtime_limit.max_background_jobs_exceeded"
                in event.get("event.reason", "")
            ]
            self.assertTrue(denied, msg=f"missing runtime-limit denial in {events}")


if __name__ == "__main__":
    unittest.main()



"""Unit tests for env_vars_files loading behavior in CheckConfig."""

import os
import tempfile
import unittest
from unittest.mock import patch, call

from lshell.checkconfig import CheckConfig

TOPDIR = f"{os.path.dirname(os.path.realpath(__file__))}/../"
CONFIG = f"{TOPDIR}/test/testfiles/test.conf"


class TestEnvVarsFilesUnit(unittest.TestCase):
    """Config-level unit tests for env_vars_files behavior."""

    args = [f"--config={CONFIG}", "--quiet=1"]

    @patch("lshell.checkconfig.builtincmd.cmd_source", return_value=0)
    def test_checkconfig_calls_cmd_source_for_each_env_file(self, mock_cmd_source):
        """Load each configured env_vars_files entry through cmd_source."""
        files = ["/tmp/a.env", "/tmp/b.env"]
        CheckConfig(self.args + [f"--env_vars_files={files}"]).returnconf()
        self.assertEqual(
            mock_cmd_source.call_args_list,
            [call("/tmp/a.env"), call("/tmp/b.env")],
        )

    def test_env_vars_files_exports_override_env_vars(self):
        """Load env_vars first, then let env_vars_files override collisions."""
        key = "LSHELL_ENV_VARS_FILES_OVERRIDE"
        prev = os.environ.get(key)

        with tempfile.NamedTemporaryFile("w", delete=False) as env_file:
            env_file.write(f"export {key}=from_file\n")
            env_path = env_file.name

        try:
            CheckConfig(
                self.args
                + [
                    f"--env_vars={{'{key}':'from_conf'}}",
                    f"--env_vars_files=['{env_path}']",
                ]
            ).returnconf()
            self.assertEqual(os.environ.get(key), "from_file")
        finally:
            os.remove(env_path)
            if prev is None:
                os.environ.pop(key, None)
            else:
                os.environ[key] = prev


if __name__ == "__main__":
    unittest.main()



"""Functional integration tests for lshell setup-system command."""

import contextlib
import grp
import io
import os
import pwd
import stat
import tempfile
import unittest
from unittest.mock import patch

from lshell import systemsetup


class TestSystemSetupFunctional(unittest.TestCase):
    """Exercise setup-system with real filesystem side effects."""

    def _current_user_and_group(self):
        """Return current account names for owner/group CLI flags."""
        username = pwd.getpwuid(os.getuid()).pw_name
        group_name = grp.getgrgid(os.getgid()).gr_name
        return username, group_name

    def test_ensure_shell_entry_is_idempotent_with_override_file(self):
        """Register shell path exactly once when called repeatedly."""
        with tempfile.TemporaryDirectory(prefix="lshell-setup-shells-") as tempdir:
            shells_file = os.path.join(tempdir, "shells")
            shell_path = "/usr/local/bin/lshell"
            with open(shells_file, "w", encoding="utf-8") as handle:
                handle.write("/bin/sh\n")

            with patch.dict(
                os.environ, {"LSHELL_SHELLS_FILE": shells_file}, clear=False
            ):
                systemsetup._ensure_shell_entry(shell_path)
                systemsetup._ensure_shell_entry(shell_path)

            with open(shells_file, "r", encoding="utf-8") as handle:
                entries = [line.strip() for line in handle if line.strip()]

            self.assertEqual(entries.count(shell_path), 1)

    def test_main_integration_creates_log_dir_and_registers_shell(self):
        """Run main flow and verify persisted filesystem effects."""
        with tempfile.TemporaryDirectory(prefix="lshell-setup-main-") as tempdir:
            log_dir = os.path.join(tempdir, "var", "log", "lshell")
            shells_file = os.path.join(tempdir, "shells")
            fake_shell = os.path.join(tempdir, "bin", "lshell")
            os.makedirs(os.path.dirname(fake_shell), exist_ok=True)
            with open(fake_shell, "w", encoding="utf-8") as handle:
                handle.write("#!/bin/sh\nexit 0\n")
            os.chmod(fake_shell, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)

            owner, group_name = self._current_user_and_group()
            stdout = io.StringIO()
            stderr = io.StringIO()

            with patch("lshell.systemsetup.os.geteuid", return_value=0):
                with patch("lshell.systemsetup.os.chown"):
                    with patch.dict(
                        os.environ, {"LSHELL_SHELLS_FILE": shells_file}, clear=False
                    ):
                        with contextlib.redirect_stdout(stdout):
                            with contextlib.redirect_stderr(stderr):
                                code = systemsetup.main(
                                    [
                                        "--group",
                                        group_name,
                                        "--owner",
                                        owner,
                                        "--log-dir",
                                        log_dir,
                                        "--shell-path",
                                        fake_shell,
                                        "--mode",
                                        "2770",
                                    ]
                                )

            self.assertEqual(code, 0)
            self.assertTrue(os.path.isdir(log_dir))
            mode = stat.S_IMODE(os.stat(log_dir).st_mode)
            self.assertEqual(mode & 0o770, 0o770)
            self.assertIn("lshell setup complete:", stdout.getvalue())
            self.assertEqual(stderr.getvalue(), "")

            with open(shells_file, "r", encoding="utf-8") as handle:
                shells_entries = [line.strip() for line in handle if line.strip()]
            self.assertIn(os.path.realpath(fake_shell), shells_entries)

    def test_main_integration_skip_shell_registration(self):
        """Skip shell registration should not create the shells file."""
        with tempfile.TemporaryDirectory(prefix="lshell-setup-skip-shells-") as tempdir:
            log_dir = os.path.join(tempdir, "var", "log", "lshell")
            shells_file = os.path.join(tempdir, "shells")
            fake_shell = os.path.join(tempdir, "bin", "lshell")
            os.makedirs(os.path.dirname(fake_shell), exist_ok=True)
            with open(fake_shell, "w", encoding="utf-8") as handle:
                handle.write("#!/bin/sh\nexit 0\n")
            os.chmod(fake_shell, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)

            owner, group_name = self._current_user_and_group()
            with patch("lshell.systemsetup.os.geteuid", return_value=0):
                with patch("lshell.systemsetup.os.chown"):
                    with patch.dict(
                        os.environ, {"LSHELL_SHELLS_FILE": shells_file}, clear=False
                    ):
                        code = systemsetup.main(
                            [
                                "--group",
                                group_name,
                                "--owner",
                                owner,
                                "--log-dir",
                                log_dir,
                                "--shell-path",
                                fake_shell,
                                "--skip-shell-registration",
                            ]
                        )

            self.assertEqual(code, 0)
            self.assertFalse(os.path.exists(shells_file))


if __name__ == "__main__":
    unittest.main()



"""Functional tests for lshell for exit command"""

import os
import unittest
from getpass import getuser
import pexpect


TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
CONFIG = f"{TOPDIR}/test/testfiles/test.conf"
LSHELL = f"{TOPDIR}/bin/lshell"
USER = getuser()
PROMPT = f"{USER}:~\\$"


class TestFunctions(unittest.TestCase):
    """Functional tests for lshell"""

    def setUp(self):
        """spawn lshell with pexpect and return the child"""
        self.child = pexpect.spawn(f"{LSHELL} --config {CONFIG} --strict 1")
        self.child.expect(PROMPT)

    def tearDown(self):
        self.child.close()

    def do_exit(self, child):
        """Exit the shell"""
        child.sendline("exit")
        child.expect(pexpect.EOF)

    def test_30_disable_exit(self):
        """F31 | test disabled exit command"""
        child = pexpect.spawn(f"{LSHELL} " f"--config {CONFIG} " "--disable_exit 1 ")
        child.expect(PROMPT)

        expected = ""
        child.sendline("exit")
        child.expect(PROMPT)

        result = child.before.decode("utf8").split("\n")[1]

        self.assertIn(expected, result)

    def test_50_warnings_then_kickout(self):
        """F50 | kicked out after warning counter"""
        child = pexpect.spawn(
            f"{LSHELL} --config {CONFIG} --strict 1 --warning_counter 0"
        )
        child.sendline("lslsls")
        child.sendline("lslsls")
        child.expect(pexpect.EOF, timeout=10)

        # Assert that the process exited
        self.assertIsNotNone(
            child.exitstatus, "The lshell process did not exit as expected."
        )

        # Optionally, you can assert that the exit code is correct
        self.assertEqual(child.exitstatus, 1, "The process should exit with code 1.")
        self.do_exit(child)



"""Security-focused unit tests for SSH/SCP/SFTP execution paths."""

import io
import os
import tempfile
import unittest
from unittest.mock import patch

from lshell.checkconfig import CheckConfig
from lshell.shellcmd import ShellCmd

TOPDIR = f"{os.path.dirname(os.path.realpath(__file__))}/../"
CONFIG = f"{TOPDIR}/test/testfiles/test.conf"


class TestSSHScpSftpAttackSurface(unittest.TestCase):
    """Protocol hardening tests split from general attack surface coverage."""

    args = [f"--config={CONFIG}", "--quiet=1"]

    def _without_ssh_env(self):
        saved = {}
        for key in ("SSH_CLIENT", "SSH_TTY", "SSH_ORIGINAL_COMMAND"):
            saved[key] = os.environ.get(key)
            os.environ.pop(key, None)
        return saved

    def _restore_ssh_env(self, saved):
        for key, value in saved.items():
            if value is None:
                os.environ.pop(key, None)
            else:
                os.environ[key] = value

    def _with_forced_ssh_env(self):
        saved = {}
        for key in ("SSH_CLIENT", "SSH_TTY", "SSH_ORIGINAL_COMMAND"):
            saved[key] = os.environ.get(key)
        os.environ["SSH_CLIENT"] = "127.0.0.1 22 22"
        os.environ.pop("SSH_TTY", None)
        return saved

    def test_shell_escape_c_runs_allowed_command_when_not_over_ssh(self):
        """Allow local -c shell escape for commands already authorized by policy."""
        saved_env = self._without_ssh_env()
        try:
            conf = CheckConfig(self.args + ["--allowed=['ls']", "--strict=0"]).returnconf()
            conf["ssh"] = "ls"
            with patch("lshell.shellcmd.utils.cmd_parse_execute", return_value=0) as mock_exec:
                with self.assertRaises(SystemExit) as cm:
                    ShellCmd(
                        conf,
                        args=[],
                        stdin=io.StringIO(),
                        stdout=io.StringIO(),
                        stderr=io.StringIO(),
                    )
            self.assertEqual(cm.exception.code, 0)
            mock_exec.assert_called_once_with("ls", shell_context=unittest.mock.ANY)
        finally:
            self._restore_ssh_env(saved_env)

    def test_shell_escape_c_blocks_disallowed_command_when_not_over_ssh(self):
        """Block local -c shell escape when the command is not in allowed policy."""
        saved_env = self._without_ssh_env()
        try:
            conf = CheckConfig(self.args + ["--allowed=['ls']", "--strict=0"]).returnconf()
            conf["ssh"] = "tail /etc/passwd"
            with patch("lshell.shellcmd.utils.cmd_parse_execute") as mock_exec:
                with self.assertRaises(SystemExit) as cm:
                    ShellCmd(
                        conf,
                        args=[],
                        stdin=io.StringIO(),
                        stdout=io.StringIO(),
                        stderr=io.StringIO(),
                    )
            self.assertEqual(cm.exception.code, 1)
            mock_exec.assert_not_called()
        finally:
            self._restore_ssh_env(saved_env)

    def test_run_overssh_allows_command_present_in_overssh(self):
        """Execute forced SSH command when it is explicitly in overssh list."""
        saved_env = self._with_forced_ssh_env()
        try:
            conf = CheckConfig(
                self.args + ["--allowed=['echo']", "--overssh=['ls']", "--strict=0"]
            ).returnconf()
            conf["ssh"] = "ls"
            with patch("lshell.shellcmd.utils.cmd_parse_execute", return_value=0) as mock_exec:
                with self.assertRaises(SystemExit) as cm:
                    ShellCmd(
                        conf,
                        args=[],
                        stdin=io.StringIO(),
                        stdout=io.StringIO(),
                        stderr=io.StringIO(),
                    )
            self.assertEqual(cm.exception.code, 0)
            mock_exec.assert_called_once_with("ls", shell_context=unittest.mock.ANY)
        finally:
            self._restore_ssh_env(saved_env)

    def test_run_overssh_strips_auto_ls_alias_before_execution(self):
        """Keep the synthetic local ls alias from mutating forced SSH commands."""
        saved_env = self._with_forced_ssh_env()
        try:
            conf = CheckConfig(
                self.args + ["--allowed=['echo']", "--overssh=['ls']", "--strict=0"]
            ).returnconf()
            if not conf.get("_auto_ls_alias") or conf["aliases"].get("ls") is None:
                self.skipTest("platform does not synthesize an ls alias")

            conf["ssh"] = "ls"
            with patch("lshell.shellcmd.utils.cmd_parse_execute", return_value=0) as mock_exec:
                with self.assertRaises(SystemExit) as cm:
                    ShellCmd(
                        conf,
                        args=[],
                        stdin=io.StringIO(),
                        stdout=io.StringIO(),
                        stderr=io.StringIO(),
                    )
            self.assertEqual(cm.exception.code, 0)
            mock_exec.assert_called_once_with("ls", shell_context=unittest.mock.ANY)
        finally:
            self._restore_ssh_env(saved_env)

    def test_run_overssh_rejects_command_not_in_overssh(self):
        """Deny forced SSH command even when it is present in normal allowed list."""
        saved_env = self._with_forced_ssh_env()
        try:
            conf = CheckConfig(
                self.args + ["--allowed=['ls']", "--overssh=['echo']", "--strict=0"]
            ).returnconf()
            conf["ssh"] = "ls"
            with patch("lshell.shellcmd.utils.cmd_parse_execute") as mock_exec:
                with self.assertRaises(SystemExit) as cm:
                    ShellCmd(
                        conf,
                        args=[],
                        stdin=io.StringIO(),
                        stdout=io.StringIO(),
                        stderr=io.StringIO(),
                    )
            self.assertEqual(cm.exception.code, 1)
            mock_exec.assert_not_called()
        finally:
            self._restore_ssh_env(saved_env)

    def test_run_overssh_rejects_forbidden_chars(self):
        """Deny forced SSH command containing forbidden separators."""
        saved_env = self._with_forced_ssh_env()
        try:
            conf = CheckConfig(
                self.args + ["--allowed=['ls']", "--overssh=['ls']", "--strict=0"]
            ).returnconf()
            conf["ssh"] = "ls; echo pwned"
            with patch("lshell.shellcmd.utils.cmd_parse_execute") as mock_exec:
                with self.assertRaises(SystemExit) as cm:
                    ShellCmd(
                        conf,
                        args=[],
                        stdin=io.StringIO(),
                        stdout=io.StringIO(),
                        stderr=io.StringIO(),
                    )
            self.assertEqual(cm.exception.code, 1)
            mock_exec.assert_not_called()
        finally:
            self._restore_ssh_env(saved_env)

    def test_run_overssh_rejects_sftp_when_disabled(self):
        """Deny sftp-server sessions when sftp flag is disabled."""
        saved_env = self._with_forced_ssh_env()
        try:
            conf = CheckConfig(self.args + ["--sftp=0", "--strict=0"]).returnconf()
            conf["ssh"] = "/usr/libexec/sftp-server"
            with patch("lshell.shellcmd.utils.cmd_parse_execute") as mock_exec:
                with self.assertRaises(SystemExit) as cm:
                    ShellCmd(
                        conf,
                        args=[],
                        stdin=io.StringIO(),
                        stdout=io.StringIO(),
                        stderr=io.StringIO(),
                    )
            self.assertEqual(cm.exception.code, 1)
            mock_exec.assert_not_called()
        finally:
            self._restore_ssh_env(saved_env)

    def test_run_overssh_allows_sftp_when_enabled(self):
        """Execute sftp-server sessions when sftp flag is enabled."""
        saved_env = self._with_forced_ssh_env()
        try:
            conf = CheckConfig(self.args + ["--sftp=1", "--strict=0"]).returnconf()
            conf["ssh"] = "/usr/libexec/sftp-server"
            with patch("lshell.shellcmd.utils.cmd_parse_execute", return_value=0) as mock_exec:
                with self.assertRaises(SystemExit) as cm:
                    ShellCmd(
                        conf,
                        args=[],
                        stdin=io.StringIO(),
                        stdout=io.StringIO(),
                        stderr=io.StringIO(),
                    )
            self.assertEqual(cm.exception.code, 0)
            mock_exec.assert_called_once_with(
                "/usr/libexec/sftp-server",
                shell_context=unittest.mock.ANY,
                trusted_protocol=True,
            )
        finally:
            self._restore_ssh_env(saved_env)

    def test_run_overssh_rejects_scp_when_disabled_and_not_in_overssh(self):
        """Deny scp transfer when global scp flag is disabled."""
        saved_env = self._with_forced_ssh_env()
        try:
            conf = CheckConfig(self.args + ["--scp=0", "--overssh=[]", "--strict=0"]).returnconf()
            conf["ssh"] = f"scp -f {conf['home_path']}/artifact"
            with patch("lshell.shellcmd.utils.cmd_parse_execute") as mock_exec:
                with self.assertRaises(SystemExit) as cm:
                    ShellCmd(
                        conf,
                        args=[],
                        stdin=io.StringIO(),
                        stdout=io.StringIO(),
                        stderr=io.StringIO(),
                    )
            self.assertEqual(cm.exception.code, 1)
            mock_exec.assert_not_called()
        finally:
            self._restore_ssh_env(saved_env)

    def test_run_overssh_allows_scp_from_overssh_even_if_scp_flag_disabled(self):
        """Allow scp transfer when scp is present in overssh allowlist."""
        saved_env = self._with_forced_ssh_env()
        try:
            conf = CheckConfig(
                self.args + ["--scp=0", "--overssh=['scp']", "--scp_download=1", "--strict=0"]
            ).returnconf()
            conf["ssh"] = f"scp -f {conf['home_path']}/artifact"
            with patch("lshell.shellcmd.utils.cmd_parse_execute", return_value=0) as mock_exec:
                with self.assertRaises(SystemExit) as cm:
                    ShellCmd(
                        conf,
                        args=[],
                        stdin=io.StringIO(),
                        stdout=io.StringIO(),
                        stderr=io.StringIO(),
                    )
            self.assertEqual(cm.exception.code, 0)
            mock_exec.assert_called_once_with(
                f"scp -f {conf['home_path']}/artifact",
                shell_context=unittest.mock.ANY,
                trusted_protocol=False,
            )
        finally:
            self._restore_ssh_env(saved_env)

    def test_run_overssh_rejects_scp_download_when_scp_download_disabled(self):
        """Deny scp -f when scp_download flag is disabled."""
        saved_env = self._with_forced_ssh_env()
        try:
            conf = CheckConfig(
                self.args + ["--scp=1", "--scp_download=0", "--strict=0"]
            ).returnconf()
            conf["ssh"] = f"scp -f {conf['home_path']}/artifact"
            with patch("lshell.shellcmd.utils.cmd_parse_execute") as mock_exec:
                with self.assertRaises(SystemExit) as cm:
                    ShellCmd(
                        conf,
                        args=[],
                        stdin=io.StringIO(),
                        stdout=io.StringIO(),
                        stderr=io.StringIO(),
                    )
            self.assertEqual(cm.exception.code, 1)
            mock_exec.assert_not_called()
        finally:
            self._restore_ssh_env(saved_env)

    def test_run_overssh_rejects_scp_upload_when_scp_upload_disabled(self):
        """Deny scp -t when scp_upload flag is disabled."""
        saved_env = self._with_forced_ssh_env()
        try:
            conf = CheckConfig(self.args + ["--scp=1", "--scp_upload=0", "--strict=0"]).returnconf()
            conf["ssh"] = f"scp -t {conf['home_path']}"
            with patch("lshell.shellcmd.utils.cmd_parse_execute") as mock_exec:
                with self.assertRaises(SystemExit) as cm:
                    ShellCmd(
                        conf,
                        args=[],
                        stdin=io.StringIO(),
                        stdout=io.StringIO(),
                        stderr=io.StringIO(),
                    )
            self.assertEqual(cm.exception.code, 1)
            mock_exec.assert_not_called()
        finally:
            self._restore_ssh_env(saved_env)

    def test_run_overssh_applies_scpforce_to_upload_target(self):
        """Rewrite scp -t target path to configured scpforce directory."""
        saved_env = self._with_forced_ssh_env()
        try:
            with tempfile.TemporaryDirectory(
                prefix="lshell_scpforce_", dir=os.environ["HOME"]
            ) as forced_dir:
                conf = CheckConfig(
                    self.args
                    + ["--scp=1", "--scp_upload=1", f"--scpforce='{forced_dir}'", "--strict=0"]
                ).returnconf()
                conf["ssh"] = f"scp -t {conf['home_path']}"
                with patch("lshell.shellcmd.utils.cmd_parse_execute", return_value=0) as mock_exec:
                    with self.assertRaises(SystemExit) as cm:
                        ShellCmd(
                            conf,
                            args=[],
                            stdin=io.StringIO(),
                            stdout=io.StringIO(),
                            stderr=io.StringIO(),
                        )
            self.assertEqual(cm.exception.code, 0)
            mock_exec.assert_called_once_with(
                f"scp -t {os.path.realpath(forced_dir)}",
                shell_context=unittest.mock.ANY,
                trusted_protocol=False,
            )
        finally:
            self._restore_ssh_env(saved_env)

    def test_run_overssh_rejects_scp_chain_with_non_protocol_segment(self):
        """Reject scp chaining when a later segment is not allowed over SSH."""
        saved_env = self._with_forced_ssh_env()
        try:
            conf = CheckConfig(self.args + ["--scp=1", "--scp_download=1"]).returnconf()
            conf["ssh"] = f"scp -f {conf['home_path']}/artifact || id"
            with patch("lshell.shellcmd.utils.cmd_parse_execute") as mock_exec:
                with self.assertRaises(SystemExit) as cm:
                    ShellCmd(
                        conf,
                        args=[],
                        stdin=io.StringIO(),
                        stdout=io.StringIO(),
                        stderr=io.StringIO(),
                    )
            self.assertEqual(cm.exception.code, 1)
            mock_exec.assert_not_called()
        finally:
            self._restore_ssh_env(saved_env)

    def test_run_overssh_applies_aliases_to_scp_commands(self):
        """Expand aliases before SCP policy checks and execution."""
        saved_env = self._with_forced_ssh_env()
        try:
            conf = CheckConfig(
                self.args
                + [
                    "--scp=1",
                    "--scp_download=1",
                    f"--aliases={{'getscp':'scp -f {os.environ['HOME']}/artifact'}}",
                ]
            ).returnconf()
            conf["ssh"] = "getscp"
            with patch("lshell.shellcmd.utils.cmd_parse_execute", return_value=0) as mock_exec:
                with self.assertRaises(SystemExit) as cm:
                    ShellCmd(
                        conf,
                        args=[],
                        stdin=io.StringIO(),
                        stdout=io.StringIO(),
                        stderr=io.StringIO(),
                    )
            self.assertEqual(cm.exception.code, 0)
            mock_exec.assert_called_once_with(
                f"scp -f {conf['home_path']}/artifact",
                shell_context=unittest.mock.ANY,
                trusted_protocol=False,
            )
        finally:
            self._restore_ssh_env(saved_env)


if __name__ == "__main__":
    unittest.main()



"""Structured security audit logging helpers."""

import json
import logging
import os
from datetime import datetime, timezone


ECS_VERSION = "8.11.0"
LAST_REASON_KEY = "_last_security_decision_reason"


def _now_utc_iso():
    """Return UTC ISO8601 timestamp with millisecond precision."""
    return (
        datetime.now(timezone.utc)
        .isoformat(timespec="milliseconds")
        .replace("+00:00", "Z")
    )


def _source_ip():
    """Extract source IP from SSH context, falling back to localhost."""
    if os.environ.get("SSH_CLIENT"):
        return os.environ["SSH_CLIENT"].split()[0]
    if os.environ.get("SSH_CONNECTION"):
        return os.environ["SSH_CONNECTION"].split()[0]
    if os.environ.get("REMOTE_ADDR"):
        return os.environ["REMOTE_ADDR"]
    return "127.0.0.1"


def enabled(conf):
    """Return True when structured audit logging is enabled."""
    return bool(conf.get("security_audit_json") and conf.get("logpath"))


class EcsJsonFormatter(logging.Formatter):
    """Format log records as ECS-aligned JSON lines."""

    def format(self, record):
        payload = {
            "@timestamp": _now_utc_iso(),
            "ecs.version": ECS_VERSION,
            "log.level": record.levelname.lower(),
            "message": record.getMessage(),
            "session.id": str(
                getattr(record, "session_id", "")
                or os.environ.get("LSHELL_SESSION_ID", "")
            ),
            "source.ip": str(getattr(record, "source_ip", "") or _source_ip()),
            "user.name": str(
                getattr(record, "username", "")
                or os.environ.get("LOGNAME")
                or os.environ.get("USER")
                or ""
            ),
        }

        event_kind = getattr(record, "event_kind", None)
        if event_kind:
            payload["event.kind"] = event_kind
        event_category = getattr(record, "event_category", None)
        if event_category:
            payload["event.category"] = event_category
        event_type = getattr(record, "event_type", None)
        if event_type:
            payload["event.type"] = event_type
        event_action = getattr(record, "event_action", None)
        if event_action:
            payload["event.action"] = event_action
        event_outcome = getattr(record, "event_outcome", None)
        if event_outcome:
            payload["event.outcome"] = event_outcome
        event_reason = getattr(record, "event_reason", None)
        if event_reason:
            payload["event.reason"] = event_reason
        process_command_line = getattr(record, "process_command_line", None)
        if process_command_line:
            payload["process.command_line"] = process_command_line

        allowed = getattr(record, "lshell_security_allowed", None)
        if allowed is not None:
            payload["lshell.security.allowed"] = bool(allowed)

        return json.dumps(payload, sort_keys=True)


def set_decision_reason(conf, reason):
    """Store latest decision reason in session config."""
    conf[LAST_REASON_KEY] = reason


def pop_decision_reason(conf, default="policy evaluation failed"):
    """Return and clear latest decision reason from session config."""
    return conf.pop(LAST_REASON_KEY, default)


def log_command_event(conf, command, allowed, reason, level=None):
    """Emit one ECS-aligned command authorization event."""
    log_security_event(
        conf,
        action="command_authorization",
        allowed=allowed,
        reason=reason,
        command=command,
        level=level,
        message="lshell command authorization decision",
    )


def log_security_event(
    conf,
    action,
    allowed,
    reason,
    command="",
    level=None,
    message="lshell security decision",
):
    """Emit one ECS-aligned runtime security event."""
    if not enabled(conf):
        return

    logger = conf["logpath"]
    log_method = str(level or ("info" if allowed else "warning")).lower()
    log_level = getattr(logging, log_method.upper(), logging.INFO)
    logger.log(
        log_level,
        message,
        extra={
            "session_id": str(conf.get("session_id", "")),
            "source_ip": _source_ip(),
            "username": str(conf.get("username", "")),
            "event_kind": "event",
            "event_category": ["authentication", "process"],
            "event_type": ["access"],
            "event_action": str(action),
            "event_outcome": "success" if allowed else "failure",
            "event_reason": str(reason),
            "process_command_line": str(command or ""),
            "lshell_security_allowed": bool(allowed),
        },
    )



"""This module contains the checkconfig class of lshell"""

import sys
import os
import configparser
from getpass import getuser
import string
import re
import getopt
import logging
import grp
import time
import glob
import subprocess
from logging.handlers import SysLogHandler

# import lshell specifics
from lshell import utils
from lshell import variables
from lshell import builtincmd
from lshell import configschema
from lshell import audit
from lshell import containment


class CheckConfig:
    """Check the configuration file."""

    def noexec_library_usable(self, path_noexec):
        """Return True when a noexec library can be safely preloaded."""
        probe_env = dict(os.environ)
        probe_env["LD_PRELOAD"] = path_noexec
        probe_env.pop("BASH_ENV", None)
        probe_env.pop("ENV", None)

        try:
            probe = subprocess.run(
                ["bash", "-c", "/usr/bin/true"],
                env=probe_env,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                check=False,
            )
        except OSError:
            return False

        return probe.returncode == 0

    def __init__(self, args, refresh=None, stdin=None, stdout=None, stderr=None):
        """Force the calling of the methods below"""
        if stdin is None:
            self.stdin = sys.stdin
        else:
            self.stdin = stdin
        if stdout is None:
            self.stdout = sys.stdout
        else:
            self.stdout = stdout
        if stderr is None:
            self.stderr = sys.stderr
        else:
            self.stderr = stderr

        self.refresh = refresh
        self.conf = {}
        self.conf, self.arguments = self.getoptions(args, self.conf)
        configfile = self.conf["configfile"]
        self.check_config_file_exists(configfile)
        self.conf["config_mtime"] = self.get_config_mtime(configfile)
        self.check_config_file(configfile)
        self.get_global()
        self.check_log()
        self.check_script()
        self.get_config()
        self.check_user_integrity()
        self.get_config_user()
        self.check_env()
        self.set_noexec()

    def check_config_file_exists(self, configfile):
        """Check if the configuration file exists, else exit with error"""
        if not os.path.exists(configfile):
            self.stderr.write("lshell: config file doesn't exist\n")
            utils.usage()

    def check_script(self):
        """Check if lshell is invoked with the correct binary and script extension"""
        if sys.argv[0].endswith("bin/lshell") and sys.argv[-1].endswith(".lsh"):
            script_path = os.path.realpath(sys.argv[-1])
            self.log.debug(f"Detected script mode: {script_path}")
            if not os.path.isfile(script_path):
                self.log.error(f"Script file not found: {script_path}")
                sys.exit(1)
            else:
                self.conf["script"] = script_path

    def getoptions(self, arguments, conf):
        """This method checks the usage. lshell.py must be called with a
        configuration file.
        If no configuration file is specified, it will set the configuration
        file path to /etc/lshell.conf
        """
        # set configfile as default configuration file
        conf["configfile"] = variables.configfile

        try:
            optlist, args = getopt.getopt(arguments, "hc:", variables.configparams)
        except getopt.GetoptError:
            self.stderr.write("lshell: missing or unknown argument(s)\n")
            utils.usage()

        for option, value in optlist:
            if option in ["--config"]:
                conf["configfile"] = os.path.realpath(value)
            if option in ["--log"]:
                conf["logpath"] = os.path.realpath(value)
            if f"{option[2:]}=" in variables.configparams:
                conf[option[2:]] = value
            if option in ["-c"]:
                conf["ssh"] = value.strip()
            if option in ["-h", "--help"]:
                utils.usage(exitcode=0)
            if option in ["--version"]:
                utils.version()

        # put the expanded path of configfile and logpath (if exists) in
        # LSHELL_ARGS environment variable
        args = ["--config", conf["configfile"]]
        if "logpath" in conf:
            args += ["--log", conf["logpath"]]
        os.environ["LSHELL_ARGS"] = str(args)

        # if lshell is invoked using shh autorized_keys file e.g.
        # command="/usr/bin/lshell", ssh-dss ....
        if "SSH_ORIGINAL_COMMAND" in os.environ:
            conf["ssh"] = os.environ["SSH_ORIGINAL_COMMAND"]

        return conf, args

    def check_env(self):
        """Load environment variable set in configuration file"""
        if "env_vars" in self.conf:
            env_vars = self.conf["env_vars"]
            for key in env_vars.keys():
                os.environ[key] = str(env_vars[key])

        # Check paths to files that contain env vars
        if "env_vars_files" in self.conf:
            for envfile in self.conf["env_vars_files"]:
                builtincmd.cmd_source(envfile)

    def check_config_file(self, file):
        """This method checks the existence of the given configuration
        file passed via command line arguments
        """
        if not os.path.exists(file):
            self.stderr.write("lshell: config file doesn't exist\n")
            utils.usage()
        else:
            self.config = configparser.ConfigParser(interpolation=None)

    def get_global(self):
        """Loads the [global] parameters from the configuration file"""
        try:
            self.config.read(self.conf["configfile"])
        except (
            configparser.MissingSectionHeaderError,
            configparser.ParsingError,
        ) as argument:
            self.stderr.write(f"lshell: invalid config file: {argument}\n")
            sys.exit(1)

        if not self.config.has_section("global"):
            self.stderr.write("lshell: config file missing [global] section\n")
            sys.exit(1)

        for item in self.config.items("global"):
            if item[0] not in self.conf:
                self.conf[item[0]] = item[1]

    def check_log(self):
        """Sets the log level and log file"""
        # define log levels dict
        self.levels = {
            1: logging.CRITICAL,
            2: logging.ERROR,
            3: logging.WARNING,
            4: logging.DEBUG,
        }

        # create logger for lshell application
        if self.conf.get("syslogname"):
            try:
                logname = str(self.conf["syslogname"])
            except (SyntaxError, NameError, TypeError):
                sys.stderr.write("lshell: syslogname must be a string\n")
                sys.exit(1)
        else:
            logname = "lshell"

        logger = logging.getLogger(f"{logname}.{self.conf['config_mtime']}")

        # close any logger handler/filters if exists
        # this is useful if configuration is reloaded
        for loghandler in logger.handlers:
            try:
                logging.shutdown(loghandler)
            except TypeError:
                pass
        for logfilter in logger.filters:
            logger.removeFilter(logfilter)

        logger.setLevel(logging.DEBUG)

        # set log to output error on stderr
        logsterr = logging.StreamHandler()
        logger.addHandler(logsterr)
        logsterr.setFormatter(logging.Formatter("%(message)s"))
        logsterr.setLevel(logging.CRITICAL)

        # log level must be 1, 2, 3 , 4 or 0
        if "loglevel" not in self.conf:
            self.conf["loglevel"] = 0
        try:
            self.conf["loglevel"] = int(self.conf["loglevel"])
        except ValueError:
            self.conf["loglevel"] = 0
        if self.conf["loglevel"] > 4:
            self.conf["loglevel"] = 4
        elif self.conf["loglevel"] < 0:
            self.conf["loglevel"] = 0

        try:
            structured_audit_enabled = int(self.conf.get("security_audit_json", 0)) == 1
        except (TypeError, ValueError):
            structured_audit_enabled = False

        if structured_audit_enabled:
            formatter = audit.EcsJsonFormatter()
            syslogformatter = audit.EcsJsonFormatter()
        else:
            formatter = logging.Formatter(f"%(asctime)s ({getuser()}): %(message)s")
            syslogformatter = logging.Formatter(
                f"{logname}[{os.getpid()}]: {getuser()}: %(message)s"
            )

        # read logfilename is exists, and set logfilename
        if self.conf.get("logfilename"):
            try:
                logfilename = str(self.conf["logfilename"])
            except (SyntaxError, NameError, TypeError):
                sys.stderr.write("lshell: logfilename must be a string\n")
                sys.exit(1)
            currentime = time.localtime()
            logfilename = logfilename.replace("%y", f"{currentime[0]}")
            logfilename = logfilename.replace("%m", f"{currentime[1]:02d}")
            logfilename = logfilename.replace("%d", f"{currentime[2]:02d}")
            logfilename = logfilename.replace(
                "%h", f"{currentime[3]:02d}{currentime[4]:02d}"
            )
            logfilename = logfilename.replace("%u", getuser())
        else:
            logfilename = getuser()

        log_directory = self.conf["logpath"]

        if self.conf["loglevel"] > 0:
            try:
                if logfilename == "syslog":
                    syslog = SysLogHandler(address="/dev/log")
                    syslog.setFormatter(syslogformatter)
                    syslog.setLevel(self.levels[self.conf["loglevel"]])
                    logger.addHandler(syslog)
                else:
                    # if log file is writable add new log file handler
                    logfile = os.path.join(log_directory, logfilename + ".log")
                    # create log file if it does not exist, and set permissions
                    with open(logfile, "a", encoding="utf-8"):
                        pass
                    try:
                        # Group-writable logs support shared operational access
                        # when /var/log/lshell is managed with group ownership.
                        os.chmod(logfile, 0o660)
                    except OSError:
                        pass
                    # set logging handler
                    self.logfile = logging.FileHandler(logfile)
                    self.logfile.setFormatter(formatter)
                    self.logfile.setLevel(self.levels[self.conf["loglevel"]])
                    logger.addHandler(self.logfile)

            except IOError:
                pass

        self.conf["logpath"] = logger
        self.log = logger

    def get_config(self):
        """Load default, group and user configuration. Then merge them all.
        The load priority is done in the following order:
            1- User section
            2- Group section
            3- Default section
        """
        self.config.read(self.conf["configfile"])

        # list the include_dir directory and read configuration files
        if "include_dir" in self.conf:
            self.conf["include_dir_conf"] = glob.glob(f"{self.conf['include_dir']}*")
            self.config.read(self.conf["include_dir_conf"])

        self.user = getuser()

        self.conf_raw = {}

        # get 'default' configuration if any
        self.get_config_sub("default")

        # get groups configuration if any.
        # for each group the user belongs to, check if specific configuration
        # exists.  The primary group has the highest priority.
        grplist = os.getgroups()
        grplist.reverse()
        for gid in grplist:
            try:
                grpname = grp.getgrgid(gid)[0]
                section = "grp:" + grpname
                self.get_config_sub(section)
            except KeyError:
                pass

        # get user configuration if any
        self.get_config_sub(self.user)

    def get_config_sub(self, section):
        """this function is used to interpret the configuration +/-,
        'all' etc.
        """
        # convert commandline options from dict to list of tuples, in order to
        # merge them with the output of the config parser
        conf = []
        for key in self.conf:
            if key not in ["config_mtime", "logpath"]:
                conf.append((key, self.conf[key]))

        if self.config.has_section(section):
            conf = list(self.config.items(section)) + conf
            for item in conf:
                key = item[0]
                value = item[1]
                # if string, then split
                split = [""]
                if isinstance(value, str):
                    split = re.split(r"((?:\+|-)\s*\[[^\]]+\])", value)
                if len(split) > 1 and key in [
                    "path",
                    "overssh",
                    "allowed",
                    "allowed_shell_escape",
                    "allowed_file_extensions",
                    "forbidden",
                ]:
                    for stuff in split:
                        if not stuff.strip():
                            continue
                        if stuff.startswith("-") or stuff.startswith("+"):
                            self.conf_raw.update(
                                self.minusplus(self.conf_raw, key, stuff)
                            )
                        elif configschema.is_all_literal(stuff):
                            if key == "allowed":
                                self.conf_raw.update({key: self.expand_all()})
                            else:
                                self.log.critical(
                                    f"lshell: config: '{key}' cannot be set to 'all'"
                                )
                                sys.exit(1)
                        elif stuff and key == "path":
                            liste = ["", ""]
                            for path in self._parse_config_value(stuff, key):
                                for item in glob.glob(path):
                                    liste[0] += os.path.realpath(item) + "/|"
                            # remove double slashes
                            liste[0] = liste[0].replace("//", "/")
                            self.conf_raw.update({key: str(liste)})
                        elif stuff and isinstance(
                            self._parse_config_value(stuff, key), list
                        ):
                            self.conf_raw.update({key: stuff})
                # case allowed/sudo_commands is set to all
                elif key == "allowed" and configschema.is_all_literal(split[0]):
                    self.conf_raw.update({key: self.expand_all()})
                elif key == "allowed_shell_escape" and configschema.is_all_literal(
                    split[0]
                ):
                    self.log.critical(
                        "lshell: config: 'allowed_shell_escape' cannot be set to 'all'"
                    )
                    sys.exit(1)
                elif key == "path":
                    liste = ["", ""]
                    for path in self._parse_config_value(value, "path"):
                        for item in glob.glob(path):
                            liste[0] += os.path.realpath(item) + "/|"
                    # remove double slashes
                    liste[0] = liste[0].replace("//", "/")
                    self.conf_raw.update({key: str(liste)})
                else:
                    self.conf_raw.update(dict([item]))

    def minusplus(self, confdict, key, extra):
        """update configuration lists containing -/+ operators"""
        if key in confdict:
            liste = self._parse_config_value(confdict[key], key)
        elif key == "path":
            liste = ["", ""]
        else:
            liste = []

        sublist = self._parse_config_value(extra[1:], key)
        if extra.startswith("+"):
            if key == "path":
                for path in sublist:
                    liste[0] += os.path.realpath(path) + "/|"
            else:
                for item in sublist:
                    liste.append(item)
        elif extra.startswith("-"):
            if key == "path":
                for path in sublist:
                    liste[1] += os.path.realpath(path) + "/|"
            else:
                for item in sublist:
                    if item in liste:
                        liste.remove(item)
                    else:
                        self.log.error(
                            f"lshell: config: -['{item}'] ignored in '{key}' list."
                        )
        return {key: str(liste)}

    def expand_all(self):
        """expand allowed, if set to 'all'"""
        # initialize list to common shell built-ins
        expanded_all = [
            "bg",
            "break",
            "case",
            "cd",
            "continue",
            "eval",
            "exec",
            "exit",
            "fg",
            "if",
            "jobs",
            "kill",
            "login",
            "logout",
            "set",
            "shift",
            "stop",
            "suspend",
            "umask",
            "unset",
            "wait",
            "while",
        ]
        for directory in os.environ["PATH"].split(":"):
            if os.path.exists(directory):
                for item in os.listdir(directory):
                    if os.access(os.path.join(directory, item), os.X_OK):
                        expanded_all.append(item)
            else:
                self.log.error(
                    f'lshell: config: PATH entry "{directory}" does not exist'
                )

        return str(expanded_all)

    def _parse_config_value(self, value, key=""):
        """Safely parse config values and enforce key schema."""
        try:
            return configschema.parse_config_value(value, key)
        except ValueError as exception:
            self.log.critical(f"lshell: config: {exception}")
            sys.exit(1)

    def _parse_history_file(self):
        """Parse and validate history_file from raw config."""
        history_value = self.conf_raw["history_file"].replace(
            "%u", self.conf["username"]
        )
        if (
            isinstance(history_value, str)
            and len(history_value) >= 2
            and history_value[0] in ("'", '"')
            and history_value[-1] == history_value[0]
        ):
            parsed = self._parse_config_value(history_value, "history_file")
        else:
            parsed = history_value

        if not isinstance(parsed, str):
            self.log.error(f"lshell: config: history file error: {history_value}")
            sys.exit(1)
        return parsed

    def check_user_integrity(self):
        """This method checks if all the required fields by user are present
        for the present user.
        In case fields are missing, the user is notified and exited from lshell
        """
        for item in variables.required_config:
            if item not in self.conf_raw:
                self.log.critical(f"lshell: missing parameter '{item}'")
                self.log.critical(
                    f"lshell: add it in the [{self.user}] or [default] section of config file."
                )
                sys.exit(1)

    def get_config_user(self):
        """Once all the checks above have passed, the configuration files
        values are entered in a dict to be used by the command line it self.
        The lshell command line is then launched!
        """
        # first, check user's loglevel
        if "loglevel" in self.conf_raw:
            try:
                self.conf["loglevel"] = int(self.conf_raw["loglevel"])
            except ValueError:
                pass
            if self.conf["loglevel"] > 4:
                self.conf["loglevel"] = 4
            elif self.conf["loglevel"] < 0:
                self.conf["loglevel"] = 0

            # if log file exists:
            try:
                self.logfile.setLevel(self.levels[self.conf["loglevel"]])
            except AttributeError:
                pass

        for item in [
            "allowed",
            "allowed_shell_escape",
            "allowed_file_extensions",
            "forbidden",
            "sudo_commands",
            "warning_counter",
            "env_vars",
            "env_vars_files",
            "timer",
            "scp",
            "scp_upload",
            "scp_download",
            "sftp",
            "overssh",
            "strict",
            "aliases",
            "messages",
            "prompt",
            "prompt_short",
            "allowed_cmd_path",
            "history_size",
            "login_script",
            "winscp",
            "disable_exit",
            "policy_commands",
            "quiet",
            "security_audit_json",
            "max_sessions_per_user",
            "max_background_jobs",
            "command_timeout",
            "max_processes",
        ]:
            try:
                if len(self.conf_raw[item]) == 0:
                    self.conf[item] = ""
                else:
                    self.conf[item] = self._parse_config_value(
                        self.conf_raw[item], item
                    )
            except KeyError:
                if item in [
                    "allowed",
                    "allowed_shell_escape",
                    "allowed_file_extensions",
                    "overssh",
                    "sudo_commands",
                    "env_vars_files",
                ]:
                    self.conf[item] = []
                elif item in ["history_size"]:
                    self.conf[item] = -1
                elif item in ["policy_commands"]:
                    self.conf[item] = 1
                # default scp is allowed
                elif item in ["scp_upload", "scp_download"]:
                    self.conf[item] = 1
                elif item in ["aliases", "env_vars", "messages"]:
                    self.conf[item] = {}
                # do not set the variable
                elif item in ["prompt"]:
                    continue
                else:
                    self.conf[item] = 0
            except TypeError:
                self.log.critical(
                    f"lshell: invalid value in '{item}'. Check the configuration file."
                )
                sys.exit(1)

        if self.conf["prompt_short"] not in [0, 1, 2]:
            self.log.critical("lshell: config: 'prompt_short' must be 0, 1, or 2")
            sys.exit(1)

        try:
            containment.validate_runtime_config(self.conf)
        except ValueError as exception:
            self.log.critical(f"lshell: config: {exception}")
            sys.exit(1)

        self.conf["username"] = self.user

        if "umask" in self.conf_raw:
            umask_raw = str(self.conf_raw["umask"]).strip().strip("'\"")
            if not re.fullmatch(r"[0-7]{1,4}", umask_raw):
                self.log.critical(
                    "lshell: config: umask must be an octal value (0000-0777), "
                    f"got: {self.conf_raw['umask']}"
                )
                sys.exit(1)
            self.conf["umask"] = umask_raw.zfill(4)
            os.umask(int(self.conf["umask"], 8))

        if "home_path" in self.conf_raw:
            home_path = self.conf_raw["home_path"]
            home_path = home_path.replace("%u", self.conf["username"])
            self.conf_raw["home_path"] = home_path
            self.conf["home_path"] = os.path.normpath(
                self._parse_config_value(self.conf_raw["home_path"], "home_path")
            )
        else:
            self.conf["home_path"] = os.environ["HOME"]

        # initialize previous path to home path
        self.conf["oldpwd"] = self.conf["home_path"]

        if "path" in self.conf_raw:
            self.conf["path"] = self._parse_config_value(self.conf_raw["path"], "path")
            self.conf["path"][0] += self.conf["home_path"]
        else:
            self.conf["path"] = ["", ""]
            self.conf["path"][0] = self.conf["home_path"]

        if "env_path" in self.conf_raw:
            self.conf["env_path"] = self._parse_config_value(
                self.conf_raw["env_path"], "env_path"
            )
        else:
            self.conf["env_path"] = ""

        if "scpforce" in self.conf_raw:
            self.conf_raw["scpforce"] = self._parse_config_value(
                self.conf_raw["scpforce"]
            )
            try:
                if os.path.exists(self.conf_raw["scpforce"]):
                    self.conf["scpforce"] = self.conf_raw["scpforce"]
                else:
                    self.log.error(
                        f"lshell: config: scpforce no such directory: {self.conf_raw['scpforce']}"
                    )
            except TypeError:
                self.log.error("lshell: config: scpforce must be a string")

        if "intro" in self.conf_raw:
            self.conf["intro"] = self._parse_config_value(self.conf_raw["intro"])
        else:
            self.conf["intro"] = variables.INTRO

        if os.path.isdir(self.conf["home_path"]):
            # change dir to home when initially loading the configuration
            if self.refresh is None:
                os.chdir(self.conf["home_path"])
            # if reloading the configuration, do not change directory
            else:
                pass
        else:
            self.log.critical(
                f'lshell: home directory "{self.conf["home_path"]}" does not exist'
            )
            sys.exit(1)

        if "history_file" in self.conf_raw:
            try:
                self.conf["history_file"] = self._parse_history_file()
            except (KeyError, TypeError):
                self.log.error(
                    f"lshell: config: history file error: {self.conf_raw.get('history_file')}"
                )
        else:
            self.conf["history_file"] = variables.HISTORY_FILE

        if not self.conf["history_file"].startswith("/"):
            self.conf["history_file"] = (
                f"{self.conf['home_path']}/{self.conf['history_file']}"
            )

        if self.conf["env_path"]:
            new_path = f"{self.conf['env_path']}:{os.environ['PATH']}"

            # Check if the new path is valid
            if all(
                c in string.ascii_letters + string.digits + "/:-_." for c in new_path
            ) and not new_path.startswith(":"):
                os.environ["PATH"] = new_path
            else:
                self.stderr.write(
                    f"lshell: config: env_path must be a valid $PATH: {self.conf['env_path']}\n"
                )
                sys.exit(1)

        # append default commands to allowed list
        self.conf["allowed"] += list(set(builtincmd.builtins_list) - set(["export"]))

        # Optionally hide policy introspection commands from users.
        if self.conf.get("policy_commands") != 1:
            self.conf["allowed"] = [
                cmd
                for cmd in self.conf["allowed"]
                if cmd not in builtincmd.POLICY_COMMANDS
            ]

        # in case sudo_commands is not empty, append sudo to allowed commands
        if self.conf["sudo_commands"]:
            self.conf["allowed"].append("sudo")

        # add all commands present in allowed_cmd_path if specified
        if self.conf["allowed_cmd_path"]:
            for path in self.conf["allowed_cmd_path"]:
                # add path to PATH env variable
                os.environ["PATH"] += f":{path}"
                # find executable file, and add them to allowed commands
                for item in os.listdir(path):
                    cmd = os.path.join(path, item)
                    if os.access(cmd, os.X_OK):
                        self.conf["allowed"].append(item)

        # case sudo_commands set to 'all', expand to all 'allowed' commands
        if "sudo_commands" in self.conf_raw and configschema.is_all_literal(
            str(self.conf_raw["sudo_commands"])
        ):
            # Keep shell-internal builtins out of sudo all-expansion while
            # preserving `ls`, which is executed via exec_cmd in lshell.
            exclude = [cmd for cmd in builtincmd.builtins_list if cmd != "ls"] + ["sudo"]
            self.conf["sudo_commands"] = list(
                dict.fromkeys(x for x in self.conf["allowed"] if x not in exclude)
            )

        # sort lsudo commands
        self.conf["sudo_commands"].sort()

        # Enable colored `ls` output by default while preserving explicit aliases.
        ls_color_alias = None
        self.conf["_auto_ls_alias"] = False
        if sys.platform.startswith("linux"):
            ls_color_alias = "ls --color=auto"
        elif sys.platform == "darwin" or "bsd" in sys.platform:
            ls_color_alias = "ls -G"

        if (
            ls_color_alias
            and "ls" in self.conf["allowed"]
            and isinstance(self.conf.get("aliases"), dict)
            and "ls" not in self.conf["aliases"]
        ):
            self.conf["aliases"]["ls"] = ls_color_alias
            self.conf["_auto_ls_alias"] = True

        # in case winscp is set, load the needed configuration
        if "winscp" in self.conf and self.conf["winscp"] == 1:
            # add minimum commands required for WinSCP to work
            self.conf["allowed"].extend(
                ["scp", "env", "pwd", "groups", "unset", "unalias"]
            )
            # WinSCP mode expects scp(1) transfers to be fully enabled.
            self.conf["scp_upload"] = 1
            self.conf["scp_download"] = 1
            # scpforce applies only to forced SSH scp uploads and is ignored
            # by WinSCP mode (which runs scp within session).
            self.conf.pop("scpforce", None)
            # remove duplicate commands, in case added in the above extension
            self.conf["allowed"] = list(set(self.conf["allowed"]))
            # allow the use of semicolon
            if ";" in self.conf["forbidden"]:
                self.conf["forbidden"].remove(";")

            self.log.error("WinSCP session started")

    def set_noexec(self):
        """This method checks the existence of the sudo_noexec library."""
        # list of standard sudo_noexec.so file location
        possible_lib = [
            "/lib/sudo_noexec.so",
            "/usr/lib/sudo_noexec.so",
            "/usr/lib/sudo/sudo_noexec.so",
            "/usr/libexec/sudo_noexec.so",
            "/usr/libexec/sudo/sudo_noexec.so",
            "/usr/local/lib/sudo_noexec.so",
            "/usr/local/lib/sudo/sudo_noexec.so",
            "/usr/local/libexec/sudo_noexec.so",
            "/usr/local/libexec/sudo/sudo_noexec.so",
            "/usr/pkg/libexec/sudo_noexec.so",
            "/lib64/sudo_noexec.so",
            "/usr/lib64/sudo/sudo_noexec.so",
        ]

        # check if alternative path is set in configuration file
        if "path_noexec" in self.conf_raw:
            self.conf["path_noexec"] = self._parse_config_value(
                self.conf_raw["path_noexec"]
            )
            # if path_noexec is empty, disable LD_PRELOAD
            # /!\ this feature should be used at the administrator's own risks!
            if self.conf["path_noexec"] == "":
                return
            if not os.path.exists(self.conf["path_noexec"]):
                self.log.critical(
                    "lshell: config: 'path_noexec': "
                    f"{self.conf['path_noexec']} no such file or directory"
                )
                sys.exit(2)
        else:
            # go through the list of standard lib locations
            for path_lib in possible_lib:
                if os.path.exists(path_lib):
                    self.conf["path_noexec"] = path_lib
                    break

        # in case the library was found, set the LD_PRELOAD aliases
        if self.conf.get("path_noexec") and not self.noexec_library_usable(
            self.conf["path_noexec"]
        ):
            self.log.error(
                f"lshell: disabling incompatible noexec library: {self.conf['path_noexec']}"
            )
            self.conf.pop("path_noexec", None)

        if not self.conf.get("path_noexec"):
            # if sudo_noexec.so file is not found,  write error in log file,
            # but don't exit tp  prevent strict dependency on sudo noexec lib
            self.log.error("lshell: noexec library not found")

        self.conf["allowed"] += self.conf["allowed_shell_escape"]

    def get_config_mtime(self, configfile):
        """get configuration file modification time, and store in the
        configuration dict. This should then be used to reload the
        configuration dynamically upon file changes
        """
        return os.path.getmtime(configfile)

    def returnconf(self):
        """returns the configuration dict"""
        return self.conf



"""Unit tests for SSH/SCP/SFTP-related configuration behavior."""

import os
import tempfile
import unittest

from lshell.checkconfig import CheckConfig
from lshell import builtincmd

TOPDIR = f"{os.path.dirname(os.path.realpath(__file__))}/../"
CONFIG = f"{TOPDIR}/test/testfiles/test.conf"


class TestSSHScpSftpConfig(unittest.TestCase):
    """Config-level protocol tests split from generic unit coverage."""

    args = [f"--config={CONFIG}", "--quiet=1"]

    def test_20_winscp_allowed_commands(self):
        """U20 | when winscp is enabled, new allowed commands are automatically added."""
        args = self.args + ["--allowed=[]", "--winscp=1"]
        userconf = CheckConfig(args).returnconf()
        exclude = list(set(builtincmd.builtins_list) - set(["export"]))
        expected = exclude + ["scp", "env", "pwd", "groups", "unset", "unalias"]
        expected.sort()
        allowed = userconf["allowed"]
        allowed.sort()
        self.assertEqual(allowed, expected)

    def test_21_winscp_allowed_semicolon(self):
        """U21 | when winscp is enabled, use of semicolon is allowed."""
        args = self.args + ["--forbidden=[';']", "--winscp=1"]
        userconf = CheckConfig(args).returnconf()
        self.assertNotIn(";", userconf["forbidden"])

    def test_21b_winscp_forces_scp_transfers_enabled(self):
        """U21b | winscp should override scp_upload/scp_download to enabled."""
        args = self.args + ["--scp_upload=0", "--scp_download=0", "--winscp=1"]
        userconf = CheckConfig(args).returnconf()
        self.assertEqual(userconf["scp_upload"], 1)
        self.assertEqual(userconf["scp_download"], 1)

    def test_21c_winscp_ignores_scpforce(self):
        """U21c | winscp should ignore scpforce setting."""
        with tempfile.TemporaryDirectory() as forced_dir:
            args = self.args + [f"--scpforce='{forced_dir}'", "--winscp=1"]
            userconf = CheckConfig(args).returnconf()
            self.assertNotIn("scpforce", userconf)

    def test_21d_scp_transfer_flags_default_to_enabled(self):
        """U21d | scp_upload/scp_download default values should be enabled."""
        userconf = CheckConfig(self.args).returnconf()
        self.assertEqual(userconf["scp_upload"], 1)
        self.assertEqual(userconf["scp_download"], 1)


if __name__ == "__main__":
    unittest.main()



"""Unit tests for history_size parsing and runtime behavior."""

import io
import os
import unittest
from unittest.mock import mock_open, patch

from lshell.checkconfig import CheckConfig
from lshell.shellcmd import ShellCmd

TOPDIR = f"{os.path.dirname(os.path.realpath(__file__))}/../"
CONFIG = f"{TOPDIR}/test/testfiles/test.conf"


class TestHistorySizeUnit(unittest.TestCase):
    """Tests for config parsing and cmdloop history-size handling."""

    args = [f"--config={CONFIG}", "--quiet=1"]

    def test_history_size_defaults_to_minus_one(self):
        """Default history_size should keep readline unlimited (-1)."""
        userconf = CheckConfig(self.args).returnconf()
        self.assertEqual(userconf["history_size"], -1)

    def test_history_size_accepts_integer_override(self):
        """Parse --history_size integer values from command-line overrides."""
        userconf = CheckConfig(self.args + ["--history_size=25"]).returnconf()
        self.assertEqual(userconf["history_size"], 25)

    def test_history_size_rejects_non_integer(self):
        """Reject non-integer history_size values at config-parse time."""
        with self.assertRaises(SystemExit) as exc:
            CheckConfig(self.args + ["--history_size='abc'"]).returnconf()
        self.assertEqual(exc.exception.code, 1)

    def test_cmdloop_applies_history_size_when_history_file_exists(self):
        """Apply readline history length when history file is readable."""
        conf = CheckConfig(self.args + ["--history_size=25", "--strict=0"]).returnconf()
        shell = ShellCmd(
            conf,
            args=[],
            stdin=io.StringIO(),
            stdout=io.StringIO(),
            stderr=io.StringIO(),
        )
        shell.cmdqueue = ["exit"]

        with patch("lshell.shellcmd.readline.read_history_file") as mock_read:
            with patch("lshell.shellcmd.readline.set_history_length") as mock_len:
                with patch(
                    "lshell.shellcmd.readline.get_completer_delims",
                    return_value=" \t\n",
                ):
                    with patch("lshell.shellcmd.readline.set_completer_delims"):
                        with patch("lshell.shellcmd.readline.get_completer", return_value=None):
                            with patch("lshell.shellcmd.readline.set_completer"):
                                with patch("lshell.shellcmd.readline.parse_and_bind"):
                                    with patch("lshell.shellcmd.readline.write_history_file"):
                                        with patch(
                                            "lshell.shellcmd.sys.exit",
                                            side_effect=SystemExit,
                                        ):
                                            with self.assertRaises(SystemExit):
                                                shell.cmdloop()

        mock_read.assert_called_once_with(conf["history_file"])
        mock_len.assert_called_once_with(25)

    def test_cmdloop_applies_history_size_when_history_file_missing(self):
        """Still apply history length when history file must be created first."""
        conf = CheckConfig(self.args + ["--history_size=11", "--strict=0"]).returnconf()
        shell = ShellCmd(
            conf,
            args=[],
            stdin=io.StringIO(),
            stdout=io.StringIO(),
            stderr=io.StringIO(),
        )
        shell.cmdqueue = ["exit"]

        with patch(
            "lshell.shellcmd.readline.read_history_file",
            side_effect=[IOError(), None],
        ) as mock_read:
            with patch("lshell.shellcmd.open", mock_open()):
                with patch("lshell.shellcmd.readline.set_history_length") as mock_len:
                    with patch(
                        "lshell.shellcmd.readline.get_completer_delims",
                        return_value=" \t\n",
                    ):
                        with patch("lshell.shellcmd.readline.set_completer_delims"):
                            with patch(
                                "lshell.shellcmd.readline.get_completer",
                                return_value=None,
                            ):
                                with patch("lshell.shellcmd.readline.set_completer"):
                                    with patch("lshell.shellcmd.readline.parse_and_bind"):
                                        with patch("lshell.shellcmd.readline.write_history_file"):
                                            with patch(
                                                "lshell.shellcmd.sys.exit",
                                                side_effect=SystemExit,
                                            ):
                                                with self.assertRaises(SystemExit):
                                                    shell.cmdloop()

        self.assertEqual(mock_read.call_count, 2)
        mock_len.assert_called_once_with(11)


if __name__ == "__main__":
    unittest.main()



"""Functional tests for runtime containment limits."""

import os
import tempfile
import unittest
from getpass import getuser

import pexpect


TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
CONFIG = f"{TOPDIR}/test/testfiles/test.conf"
LSHELL = f"{TOPDIR}/bin/lshell"
USER = getuser()
PROMPT = f"{USER}:~\\$"


class TestRuntimeContainmentFunctional(unittest.TestCase):
    """Validate runtime containment limits in a real shell session."""

    def _spawn_shell(self, extra_args, env=None, timeout=15):
        command = f"{LSHELL} --config {CONFIG} {extra_args}"
        return pexpect.spawn(command, encoding="utf-8", timeout=timeout, env=env)

    def _env_with(self, **extra):
        env = os.environ.copy()
        env.update(extra)
        return env

    def _safe_exit(self, child):
        if not child.isalive():
            return
        child.sendline("exit")
        try:
            child.expect(pexpect.EOF, timeout=3)
            return
        except pexpect.TIMEOUT:
            pass

        if child.isalive():
            child.sendline("exit")
            child.expect(pexpect.EOF, timeout=5)

    def test_max_sessions_per_user_enforced(self):
        """Deny second concurrent shell when max_sessions_per_user is exceeded."""
        with tempfile.TemporaryDirectory(prefix="lshell-session-func-") as session_dir:
            env = self._env_with(LSHELL_SESSION_DIR=session_dir)
            first = self._spawn_shell("--max_sessions_per_user 1 --strict 1", env=env)
            second = None
            try:
                first.expect(PROMPT)

                second = self._spawn_shell(
                    "--max_sessions_per_user 1 --strict 1",
                    env=env,
                )
                second.expect(pexpect.EOF)
                output = second.before
                self.assertIn("session denied", output)
                self.assertIn("max_sessions_per_user=1", output)
            finally:
                if second is not None:
                    second.close(force=True)
                self._safe_exit(first)

    def test_max_background_jobs_enforced(self):
        """Deny new background command when max_background_jobs is reached."""
        child = self._spawn_shell(
            "--strict 1 --forbidden \"[]\" --allowed \"['sleep']\" "
            "--max_background_jobs 1"
        )
        try:
            child.expect(PROMPT)
            child.sendline("sleep 60 &")
            child.expect(PROMPT)

            child.sendline("sleep 60 &")
            child.expect(PROMPT)
            output = child.before
            self.assertIn("background job denied", output)
            self.assertIn("max_background_jobs=1", output)
        finally:
            self._safe_exit(child)

    def test_command_timeout_kills_long_running_command(self):
        """Terminate foreground command when command_timeout is exceeded."""
        child = self._spawn_shell(
            "--strict 1 --forbidden \"[]\" --allowed \"['sleep']\" "
            "--command_timeout 1"
        )
        try:
            child.expect(PROMPT)
            child.sendline("sleep 3")
            child.expect(PROMPT)
            output = child.before
            self.assertIn("command timed out after 1s", output)
        finally:
            self._safe_exit(child)

    def _last_non_empty_line(self, text):
        lines = [line.strip() for line in text.splitlines() if line.strip()]
        return lines[-1] if lines else ""

    def test_max_processes_denies_forking_pipeline(self):
        """Low max_processes should fail a command requiring child process forks."""
        child = self._spawn_shell(
            "--strict 1 --forbidden \"[]\" --allowed \"['cat','echo']\" "
            "--max_processes 1 --command_timeout 2"
        )
        try:
            child.expect(PROMPT)
            child.sendline("echo hi | cat")
            child.expect(PROMPT)

            child.sendline("echo $?")
            child.expect(PROMPT)
            exit_line = self._last_non_empty_line(child.before)
            self.assertNotEqual(exit_line, "0")
        finally:
            self._safe_exit(child)


if __name__ == "__main__":
    unittest.main()



"""Generate hardened baseline lshell configurations."""

import argparse
import configparser
import os
import re
import sys
from datetime import datetime, timezone

from lshell import configschema


SAFE_FORBIDDEN_OPERATORS = [";", "&", "|", "`", ">", "<", "$(", "${"]
REQUIRED_PROFILE_KEYS = {
    "allowed",
    "allowed_shell_escape",
    "forbidden",
    "warning_counter",
    "strict",
    "scp",
    "scp_upload",
    "scp_download",
    "sftp",
    "overssh",
}
UNSAFE_ALLOWED_SHELL_ESCAPE = {
    "awk",
    "bash",
    "dash",
    "env",
    "expect",
    "find",
    "fish",
    "ksh",
    "less",
    "lua",
    "man",
    "more",
    "nano",
    "nawk",
    "node",
    "perl",
    "php",
    "python",
    "python3",
    "ruby",
    "sed",
    "sh",
    "ssh",
    "vi",
    "vim",
    "xargs",
    "zsh",
}


PROFILE_DEFINITIONS = {
    "sftp-only": {
        "description": "SFTP transport only with no interactive remote command execution.",
        "global": {
            "logpath": "/var/log/lshell/",
            "loglevel": 2,
            "security_audit_json": 1,
        },
        "default": {
            "allowed": ["pwd", "ls", "cd"],
            "allowed_shell_escape": [],
            "forbidden": list(SAFE_FORBIDDEN_OPERATORS),
            "warning_counter": 2,
            "strict": 1,
            "scp": 0,
            "scp_upload": 0,
            "scp_download": 0,
            "sftp": 1,
            "overssh": [],
            "sudo_commands": [],
            "allowed_file_extensions": [],
            "path": [],
            "umask": "0077",
        },
        "explain": [
            "Use this for file-drop or file-retrieval accounts over SFTP.",
            "Interactive shell functionality is intentionally minimal.",
            "SCP is disabled to reduce protocol surface area.",
        ],
    },
    "rsync-backup": {
        "description": "Rsync over SSH for backup jobs with strict command controls.",
        "global": {
            "logpath": "/var/log/lshell/",
            "loglevel": 2,
            "security_audit_json": 1,
        },
        "default": {
            "allowed": ["rsync", "pwd"],
            "allowed_shell_escape": [],
            "forbidden": list(SAFE_FORBIDDEN_OPERATORS),
            "warning_counter": 2,
            "strict": 1,
            "scp": 0,
            "scp_upload": 0,
            "scp_download": 0,
            "sftp": 0,
            "overssh": ["rsync"],
            "sudo_commands": [],
            "allowed_file_extensions": [],
            "path": [],
            "umask": "0077",
        },
        "explain": [
            "Designed for backup endpoints that need rsync and little else.",
            "over-SSH commands are narrowed to rsync only.",
            "Strict mode enforces warning counter decrement on unknown syntax.",
        ],
    },
    "deploy-minimal": {
        "description": "Minimal deployment account for artifact sync and service rollout.",
        "global": {
            "logpath": "/var/log/lshell/",
            "loglevel": 2,
            "security_audit_json": 1,
        },
        "default": {
            "allowed": [
                "cat",
                "cp",
                "git",
                "ln",
                "ls",
                "mkdir",
                "mv",
                "pwd",
                "rsync",
                "scp",
                "tail",
                "touch",
            ],
            "allowed_shell_escape": [],
            "forbidden": list(SAFE_FORBIDDEN_OPERATORS),
            "warning_counter": 2,
            "strict": 1,
            "scp": 1,
            "scp_upload": 1,
            "scp_download": 0,
            "sftp": 0,
            "overssh": ["rsync", "scp"],
            "sudo_commands": [],
            "allowed_file_extensions": [".log", ".txt", ".yml", ".yaml"],
            "path": [],
            "umask": "0027",
        },
        "explain": [
            "Use when CI/CD or operators need tightly scoped deployment operations.",
            "SCP upload is enabled for controlled artifact delivery; download is disabled.",
            "Keep sudo disabled by default and elevate only in explicit user sections.",
        ],
    },
    "readonly-support": {
        "description": "Read-only troubleshooting profile for support and incident triage.",
        "global": {
            "logpath": "/var/log/lshell/",
            "loglevel": 2,
            "security_audit_json": 1,
        },
        "default": {
            "allowed": ["cat", "grep", "head", "ls", "pwd", "tail"],
            "allowed_shell_escape": [],
            "forbidden": list(SAFE_FORBIDDEN_OPERATORS),
            "warning_counter": 2,
            "strict": 1,
            "scp": 0,
            "scp_upload": 0,
            "scp_download": 0,
            "sftp": 0,
            "overssh": [],
            "sudo_commands": [],
            "allowed_file_extensions": [".conf", ".ini", ".json", ".log", ".txt", ".yaml"],
            "path": [],
            "umask": "0077",
        },
        "explain": [
            "Purpose-built for support users who should inspect but not modify systems.",
            "No SSH transfer protocols are enabled.",
            "allowed_file_extensions helps prevent access to unexpected file types.",
        ],
    },
}

FIELD_COMMENTS = {
    "allowed": "Explicit allow-list only; never use 'all' for hardened baselines.",
    "allowed_shell_escape": (
        "Commands here bypass noexec restrictions; keep this list empty unless "
        "you have a reviewed exception."
    ),
    "forbidden": "Deny shell control operators that enable chaining, redirection, or substitution.",
    "warning_counter": "Session is terminated after repeated violations. -1 disables termination.",
    "strict": "Strict mode treats unknown syntax as policy violations (recommended: 1).",
    "scp": "Enable/disable SCP protocol surface.",
    "scp_upload": "Allow SCP uploads only when operationally required.",
    "scp_download": "Allow SCP downloads only when operationally required.",
    "sftp": "Enable/disable SFTP protocol surface.",
    "overssh": "Commands allowed for direct SSH command execution; keep as small as possible.",
    "sudo_commands": "Keep empty by default. Add only audited, non-interactive commands.",
    "allowed_file_extensions": (
        "Optional file-type allow-list; empty means no extension restriction."
    ),
    "path": "Optional path restrictions; empty list disables path ACL enforcement.",
    "umask": "Conservative process umask for files created in shell sessions.",
    "logpath": "Centralized log directory used by lshell.",
    "loglevel": "Logging verbosity (0-4). Use >=2 for security operations.",
    "security_audit_json": "Enable JSON/ECS audit events for SIEM ingestion.",
}


def _format_value(value):
    if isinstance(value, str):
        return value if value.isdigit() else repr(value)
    return repr(value)


def list_templates():
    """Return template rows for display."""
    rows = []
    for name in sorted(PROFILE_DEFINITIONS):
        rows.append((name, PROFILE_DEFINITIONS[name]["description"]))
    return rows


def get_profile(name):
    """Fetch profile data by name or raise ValueError."""
    try:
        return PROFILE_DEFINITIONS[name]
    except KeyError as exception:
        available = ", ".join(sorted(PROFILE_DEFINITIONS))
        raise ValueError(
            f"Unknown profile '{name}'. Available profiles: {available}"
        ) from exception


def validate_profile(profile_name, profile_data):
    """Validate profile structure and security constraints."""
    errors = []
    default = profile_data.get("default", {})
    missing = sorted(REQUIRED_PROFILE_KEYS - set(default.keys()))
    if missing:
        errors.append(f"missing required keys in [default]: {', '.join(missing)}")

    allowed_value = default.get("allowed", [])
    if configschema.is_all_literal(allowed_value) or allowed_value == "all":
        errors.append("allowed must be an explicit list and cannot be 'all'")
    if not isinstance(allowed_value, list) or not allowed_value:
        errors.append("allowed must be a non-empty list")

    strict_value = default.get("strict")
    if strict_value != 1:
        errors.append("strict must be set to 1 for hardened profiles")

    forbidden_value = default.get("forbidden", [])
    if not isinstance(forbidden_value, list):
        errors.append("forbidden must be a list")
    else:
        missing_ops = [item for item in SAFE_FORBIDDEN_OPERATORS if item not in forbidden_value]
        if missing_ops:
            errors.append(
                "forbidden list is missing hardened operators: " + ", ".join(missing_ops)
            )

    shell_escape_value = default.get("allowed_shell_escape", [])
    if shell_escape_value == "all" or configschema.is_all_literal(shell_escape_value):
        errors.append("allowed_shell_escape cannot be 'all'")
    if not isinstance(shell_escape_value, list):
        errors.append("allowed_shell_escape must be a list")
    else:
        for raw_command in shell_escape_value:
            base_command = raw_command.strip().split(" ", maxsplit=1)[0].lower()
            if base_command in UNSAFE_ALLOWED_SHELL_ESCAPE:
                errors.append(
                    "allowed_shell_escape contains unsafe command: " + raw_command
                )

    for key in ("scp", "scp_upload", "scp_download", "sftp", "warning_counter"):
        value = default.get(key)
        if not isinstance(value, int):
            errors.append(f"{key} must be an integer")

    for key in ("overssh", "sudo_commands", "allowed_file_extensions", "path"):
        value = default.get(key)
        if not isinstance(value, list):
            errors.append(f"{key} must be a list")

    if profile_name == "sftp-only":
        if default.get("sftp") != 1 or default.get("scp") != 0:
            errors.append("sftp-only profile must set sftp=1 and scp=0")
        if default.get("overssh"):
            errors.append("sftp-only profile must keep overssh empty")
    elif profile_name == "rsync-backup":
        if "rsync" not in default.get("allowed", []):
            errors.append("rsync-backup profile must allow rsync")
        if default.get("overssh") != ["rsync"]:
            errors.append("rsync-backup profile must set overssh to ['rsync']")
    elif profile_name == "readonly-support":
        if any(default.get(key) != 0 for key in ("scp", "sftp", "scp_upload", "scp_download")):
            errors.append("readonly-support profile must disable SCP and SFTP")

    return errors


def run_sanity_checks(rendered_config):
    """Sanity-parse generated config and return (ok, details)."""
    return run_sanity_checks_for_targets(rendered_config, [])


def run_sanity_checks_for_targets(rendered_config, target_sections):
    """Sanity-parse generated config and validate required section keys."""
    details = []
    parser = configparser.ConfigParser(interpolation=None)
    try:
        parser.read_string(rendered_config)
    except (configparser.Error, OSError) as exception:
        return False, [f"parser: fail ({exception})"]
    details.append("parser: pass")

    if not parser.has_section("global"):
        details.append("global-section: fail (missing [global])")
        return False, details
    details.append("global-section: pass")

    if not target_sections:
        if not parser.has_section("default"):
            details.append("default-section: fail (missing [default])")
            return False, details
        details.append("default-section: pass")
        target_sections = ["default"]
    else:
        for section in target_sections:
            if not parser.has_section(section):
                details.append(f"{section}: fail (missing section)")
                return False, details
            details.append(f"{section}: pass")

    for section in target_sections:
        for key in REQUIRED_PROFILE_KEYS:
            if not parser.has_option(section, key):
                details.append(f"{section}.{key}: fail (missing)")
                return False, details
            raw_value = parser.get(section, key)
            try:
                configschema.parse_config_value(raw_value, key)
                details.append(f"{section}.{key}: pass")
            except ValueError as exception:
                details.append(f"{section}.{key}: fail ({exception})")
                return False, details

    return True, details


def _render_section(lines, section_name, values):
    lines.extend(["", f"[{section_name}]"])
    for key, value in values.items():
        comment = FIELD_COMMENTS.get(key)
        if comment:
            lines.append(f"# {comment}")
        lines.append(f"{key:<15} : {_format_value(value)}")


def render_profile(profile_name, profile_data, groups=None, users=None):
    """Render profile as lshell configuration text with inline comments."""
    groups = groups or []
    users = users or []
    timestamp = datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%SZ")
    lines = [
        "# Generated by: lshell harden-init",
        f"# Profile: {profile_name}",
        f"# Generated (UTC): {timestamp}",
        "# Review this baseline and scope per-user/group overrides as needed.",
        "",
        "[global]",
    ]

    for key, value in profile_data["global"].items():
        comment = FIELD_COMMENTS.get(key)
        if comment:
            lines.append(f"# {comment}")
        lines.append(f"{key:<15} : {_format_value(value)}")

    if groups or users:
        lines.append("# Scoped profile sections generated from CLI target flags.")
        lines.append("# No [default] section is emitted to avoid global policy impact.")
        for group_name in groups:
            _render_section(lines, f"grp:{group_name}", profile_data["default"])
        for user_name in users:
            _render_section(lines, f"user:{user_name}", profile_data["default"])
    else:
        _render_section(lines, "default", profile_data["default"])

    lines.append("")
    return "\n".join(lines)


def explain_profile(profile_name, profile_data):
    """Return human-readable hardening rationale for a profile."""
    lines = [f"Profile: {profile_name}", f"Purpose: {profile_data['description']}", "Controls:"]
    lines.extend(f"- {item}" for item in profile_data["explain"])
    return "\n".join(lines)


def _print_sanity_checks(details, ok):
    for item in details:
        print(f"sanity: {item}")
    print("sanity: overall: pass" if ok else "sanity: overall: fail")


def _write_output(path, rendered_config):
    directory = os.path.dirname(path)
    if directory:
        os.makedirs(directory, exist_ok=True)
    with open(path, "w", encoding="utf-8") as handle:
        handle.write(rendered_config)


def _default_output_path(profile_name):
    return f"/etc/lshell.d/{profile_name}.conf"


def _run_wizard():
    print("lshell harden-init wizard")
    print("Select a hardened profile:")
    rows = list_templates()
    for index, row in enumerate(rows, start=1):
        print(f"  {index}. {row[0]} - {row[1]}")

    selection = input("Profile name or number: ").strip()
    if selection.isdigit():
        value = int(selection)
        if value < 1 or value > len(rows):
            print("lshell harden-init: invalid profile selection", file=sys.stderr)
            return 1
        profile_name = rows[value - 1][0]
    else:
        profile_name = selection

    try:
        profile_data = get_profile(profile_name)
    except ValueError as exception:
        print(f"lshell harden-init: {exception}", file=sys.stderr)
        return 1

    output_path = input(
        f"Output file path [{_default_output_path(profile_name)}]: "
    ).strip() or _default_output_path(profile_name)
    group_name = input("Optional target group (blank for none): ").strip()
    user_name = input("Optional target user (blank for none): ").strip()
    groups = [group_name] if group_name else []
    users = [user_name] if user_name else []

    rendered = render_profile(profile_name, profile_data, groups=groups, users=users)
    errors = validate_profile(profile_name, profile_data)
    if errors:
        for error in errors:
            print(f"lshell harden-init: validation failed: {error}", file=sys.stderr)
        return 1

    target_sections = [f"grp:{item}" for item in groups] + [f"user:{item}" for item in users]
    ok, details = run_sanity_checks_for_targets(rendered, target_sections)
    _print_sanity_checks(details, ok)
    if not ok:
        return 1

    _write_output(output_path, rendered)
    print(f"lshell harden-init: wrote {output_path}")
    return 0


def build_parser():
    """Build argparse parser for harden-init mode."""
    parser = argparse.ArgumentParser(
        prog="lshell harden-init",
        description="Generate secure-by-default lshell policy baseline profiles.",
    )
    parser.add_argument(
        "--list-templates",
        action="store_true",
        help="List available hardened templates.",
    )
    parser.add_argument(
        "--profile",
        choices=sorted(PROFILE_DEFINITIONS),
        help="Template/profile to render.",
    )
    parser.add_argument(
        "--output",
        help="Write rendered config to this path.",
    )
    parser.add_argument(
        "--stdout",
        action="store_true",
        help="Write rendered config to standard output.",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Render config and run sanity checks without writing files.",
    )
    parser.add_argument(
        "--explain",
        action="store_true",
        help="Print profile hardening rationale.",
    )
    parser.add_argument(
        "--group",
        action="append",
        default=[],
        help="Generate a scoped [grp:<name>] section (repeatable).",
    )
    parser.add_argument(
        "--user",
        action="append",
        default=[],
        help="Generate a scoped [user:<name>] section (repeatable).",
    )
    return parser


def _validate_target_names(label, values):
    pattern = re.compile(r"^[A-Za-z0-9_.-]+$")
    errors = []
    for item in values:
        if not pattern.match(item):
            errors.append(
                f"invalid {label} name '{item}'. Use letters, digits, underscore, dot, or dash."
            )
    return errors


def main(argv=None):
    """Entry point for `lshell harden-init`."""
    parser = build_parser()
    args = parser.parse_args(argv)

    if args.list_templates:
        if args.group or args.user:
            parser.error("--group/--user cannot be used with --list-templates")
        for name, description in list_templates():
            print(f"{name:<18} {description}")
        return 0

    if not args.profile:
        if sys.stdin.isatty() and sys.stdout.isatty():
            return _run_wizard()
        parser.error("missing --profile. Use --list-templates to discover options.")

    if args.output and args.stdout:
        parser.error("--output and --stdout cannot be used together")
    if not args.output and not args.stdout and not args.dry_run:
        args.output = _default_output_path(args.profile)

    profile_name = args.profile
    profile_data = get_profile(profile_name)
    target_errors = []
    target_errors.extend(_validate_target_names("group", args.group))
    target_errors.extend(_validate_target_names("user", args.user))
    if target_errors:
        for error in target_errors:
            print(f"lshell harden-init: {error}", file=sys.stderr)
        return 1

    validation_errors = validate_profile(profile_name, profile_data)
    if validation_errors:
        for error in validation_errors:
            print(f"lshell harden-init: validation failed: {error}", file=sys.stderr)
        return 1

    rendered = render_profile(
        profile_name, profile_data, groups=args.group, users=args.user
    )
    target_sections = [f"grp:{item}" for item in args.group] + [
        f"user:{item}" for item in args.user
    ]
    ok, details = run_sanity_checks_for_targets(rendered, target_sections)
    if args.dry_run:
        print(rendered.rstrip())
        _print_sanity_checks(details, ok)
        if args.explain:
            print("")
            print(explain_profile(profile_name, profile_data))
        return 0 if ok else 1

    if not ok:
        _print_sanity_checks(details, ok)
        return 1

    if args.explain:
        print(explain_profile(profile_name, profile_data))

    if args.stdout:
        print(rendered.rstrip())
        return 0

    try:
        _write_output(args.output, rendered)
    except OSError as exception:
        print(f"lshell harden-init: {exception}", file=sys.stderr)
        return 1

    print(f"lshell harden-init: wrote {args.output}")
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



""" This module contains the built-in commands of lshell """

import glob
import sys
import os
import re
import shlex
import readline
import signal

# import lshell specifics
from lshell import variables
from lshell import utils


# Store background jobs
BACKGROUND_JOBS = []

POLICY_COMMANDS = [
    "policy-show",
    "policy-path",
    "policy-sudo",
    "lpath",
    "lsudo",
]

builtins_list = [
    "cd",
    "ls",
    "clear",
    "exit",
    "export",
    "history",
    "policy-show",
    "policy-path",
    "policy-sudo",
    "lpath",
    "lsudo",
    "help",
    "fg",
    "bg",
    "jobs",
    "source",
]


def _cancel_job_timeout(job):
    """Cancel a watchdog timer attached to a background job, if any."""
    timer = getattr(job, "lshell_timeout_timer", None)
    if timer is not None:
        timer.cancel()


def cmd_lpath(conf):
    """Show path policy in a concise, readable format."""
    current_dir = os.path.realpath(os.getcwd())
    current_match = current_dir if current_dir.endswith("/") else f"{current_dir}/"
    allowed_re = str(conf["path"][0])
    denied_re = str(conf["path"][1][:-1])
    current_allowed = bool(re.findall(allowed_re, current_match))
    current_denied = bool(re.findall(denied_re, current_match)) if denied_re else False
    current_status = "allowed" if current_allowed and not current_denied else "denied"

    sys.stdout.write("Path Policy\n")
    sys.stdout.write("-----------\n")
    sys.stdout.write(f"Current directory      : {current_dir} ({current_status})\n")
    sys.stdout.write("\nAllowed paths\n")
    sys.stdout.write("-------------\n")
    lpath_allowed = sorted(path for path in conf["path"][0].split("|") if path)
    if lpath_allowed:
        for path in lpath_allowed:
            sys.stdout.write(f"{path}\n")
    else:
        sys.stdout.write("none\n")

    lpath_denied = sorted(path for path in conf["path"][1].split("|") if path)
    if lpath_denied:
        sys.stdout.write("\nDenied paths\n")
        sys.stdout.write("------------\n")
        for path in lpath_denied:
            sys.stdout.write(f"{path}\n")
    return 0


def cmd_lsudo(conf):
    """Show sudo policy in a concise, readable format."""
    sudo_commands = sorted(conf.get("sudo_commands", []))
    enabled = bool(sudo_commands)

    sys.stdout.write("Sudo Policy\n")
    sys.stdout.write("-----------\n")
    sys.stdout.write(f"Sudo access            : {'enabled' if enabled else 'disabled'}\n")
    if enabled:
        sys.stdout.write(f"Allowed via sudo       : {', '.join(sudo_commands)}\n")
        return 0

    sys.stdout.write("Allowed via sudo       : none\n")
    return 1


def cmd_history(conf, log):
    """print the commands history"""
    try:
        try:
            readline.write_history_file(conf["history_file"])
        except IOError:
            log.error(f"WARN: couldn't write history to file {conf['history_file']}\n")
            return 1
        with open(conf["history_file"], "r", encoding="utf-8") as history_file:
            i = 1
            for item in history_file.readlines():
                sys.stdout.write(f"{i}:  {item}")
                i += 1
    except (
        OSError,
        IOError,
        FileNotFoundError,
    ) as exception:  # Catch specific exceptions
        log.critical(f"** Unable to read the history file: {exception}")
        return 1
    return 0


def cmd_export(args):
    """export environment variables"""
    try:
        tokens = shlex.split(args, posix=True)
    except ValueError:
        return 0, None

    if len(tokens) >= 2 and "=" in tokens[1]:
        var, value = tokens[1].split("=", 1)
        # disallow dangerous variable
        if var in variables.FORBIDDEN_ENVIRON:
            return 1, var
        os.environ.update({var: value})
    return 0, None


def cmd_source(envfile):
    """Source a file in the current shell context"""
    envfile = envfile.strip().strip("'").strip('"')
    envfile = os.path.expanduser(os.path.expandvars(envfile))
    try:
        with open(envfile, encoding="utf-8") as env_vars:
            for env_var in env_vars.readlines():
                line = env_var.strip()
                if not line or line.startswith("#"):
                    continue
                if line.startswith("export "):
                    cmd_export(line)
    except (OSError, IOError):
        sys.stderr.write(f"lshell: unable to read environment file: {envfile}\n")
        return 1
    return 0


def cmd_cd(directory, conf):
    """implementation of the "cd" command"""
    # expand user's ~
    directory = os.path.expanduser(directory)

    # remove quotes if present
    directory = directory.strip("'").strip('"')

    if len(directory) >= 1:
        # add wildcard completion support to cd
        if directory.find("*"):
            # get all files and directories matching wildcard
            wildall = glob.glob(directory)
            wilddir = []
            # filter to only directories
            for item in wildall:
                if os.path.isdir(item):
                    wilddir.append(item)
            # sort results
            wilddir.sort()
            # if any results are returned, pick first one
            if len(wilddir) >= 1:
                directory = wilddir[0]
        # go previous directory
        if directory == "-":
            directory = conf["oldpwd"]

        # store current directory in oldpwd variable
        conf["oldpwd"] = os.getcwd()

        # change directory
        try:
            os.chdir(os.path.realpath(directory))
            conf["promptprint"] = utils.updateprompt(os.getcwd(), conf)
        except OSError as excp:
            sys.stdout.write(f"lshell: {directory}: {excp.strerror}\n")
            return excp.errno, conf
    else:
        os.chdir(conf["home_path"])
        conf["promptprint"] = utils.updateprompt(os.getcwd(), conf)

    return 0, conf


def check_background_jobs():
    """Check the status of background jobs and print a completion message if done."""
    active_jobs = []
    for idx, job in enumerate(BACKGROUND_JOBS, start=1):
        if job.poll() is None:
            active_jobs.append(job)
            continue

        _cancel_job_timeout(job)
        if getattr(job, "lshell_timeout_triggered", False):
            print(f"[{idx}]+  Timed Out               {_job_command(job)}")
            continue

        status = "Done" if job.returncode == 0 else "Failed"
        args = _job_command(job)
        # only print if the job has not been interrupted by the user
        if job.returncode != -2:
            print(f"[{idx}]+  {status}                    {args}")

    BACKGROUND_JOBS[:] = active_jobs


def get_job_status(job):
    """Return the status of a background job."""
    if getattr(job, "lshell_timeout_triggered", False):
        return "Timed Out"
    if job.poll() is None:
        status = "Stopped"
    elif job.poll() == 0:
        status = "Completed"  # Process completed successfully
    else:
        status = "Killed"  # Process was killed or terminated with a non-zero code
    return status


def _job_command(job):
    """Return the original command line for a tracked job."""
    return getattr(job, "lshell_cmd", " ".join(job.args))


def jobs():
    """Return a list of background jobs."""
    joblist = []
    active_jobs = []
    for job in BACKGROUND_JOBS:
        if job.poll() is not None:
            _cancel_job_timeout(job)
            continue

        active_jobs.append(job)
        idx = len(active_jobs)
        status = get_job_status(job)
        cmd = _job_command(job)
        joblist.append([idx, status, cmd])

    BACKGROUND_JOBS[:] = active_jobs
    return joblist


def cmd_jobs():
    """List all backgrounded jobs."""
    joblist = jobs()
    job_count = len(joblist)

    try:
        for i, job in enumerate(joblist, start=1):
            idx, status, cmd = job
            # Add '+' symbol for the most recent job
            job_symbol = "+"
            if job_count > 1:
                if i == job_count - 1:
                    # Add '-' for the second-to-last job
                    job_symbol = "-"
                elif i < job_count:
                    # No symbol for other jobs
                    job_symbol = " "
            print(f"[{idx}]{job_symbol}  {status}        {cmd}")
    except IndexError:
        return 1
    return 0


def cmd_bg_fg(job_type, job_id):
    """Resume a backgrounded job."""
    if job_type == "bg":
        print("lshell: bg not supported")
        return 1

    if job_id:
        # Check if job ID is valid
        try:
            job_id = int(job_id)
        except ValueError:
            print("lshell: invalid job ID")
            return 1
    else:
        # Use the last job if no specific job_id is provided
        if BACKGROUND_JOBS:
            job_id = len(BACKGROUND_JOBS)
        else:
            print(f"lshell: {job_type}: current: no such job")
            return 1

    if 0 < job_id <= len(BACKGROUND_JOBS):
        job = BACKGROUND_JOBS[job_id - 1]
        if job.poll() is None:
            if job_type == "fg":
                class CtrlZForeground(Exception):
                    """Raised when the foreground job is suspended with Ctrl+Z."""

                    pass

                def handle_sigtstp(signum, frame):
                    """Suspend the foreground job and keep/update its jobs list entry."""
                    if job.poll() is None:
                        os.killpg(os.getpgid(job.pid), signal.SIGSTOP)
                        if job in BACKGROUND_JOBS:
                            current_job_id = BACKGROUND_JOBS.index(job) + 1
                        else:
                            BACKGROUND_JOBS.append(job)
                            current_job_id = len(BACKGROUND_JOBS)
                        sys.stdout.write(
                            f"\n[{current_job_id}]+  Stopped        {_job_command(job)}\n"
                        )
                        sys.stdout.flush()
                    raise CtrlZForeground()

                previous_sigtstp_handler = signal.getsignal(signal.SIGTSTP)
                try:
                    signal.signal(signal.SIGTSTP, handle_sigtstp)
                    print(_job_command(job))
                    # Bring it to the foreground and wait
                    os.killpg(os.getpgid(job.pid), signal.SIGCONT)
                    job.wait()
                    # Remove the job from the list if it has completed
                    if job.poll() is not None:
                        BACKGROUND_JOBS.pop(job_id - 1)
                    return 0
                except CtrlZForeground:
                    return 0
                except KeyboardInterrupt:
                    os.killpg(os.getpgid(job.pid), signal.SIGINT)
                    BACKGROUND_JOBS.pop(job_id - 1)
                    return 130
                finally:
                    signal.signal(signal.SIGTSTP, previous_sigtstp_handler)
            # bg not supported at the moment
            # elif job_type == "bg":
            #     print(f"lshell: bg not supported")
            #     return 1
        else:
            print(f"lshell: {job_type}: {job_id}: no such job")
            return 1
    else:
        print(f"lshell: {job_type}: {job_id}: no such job")
        return 1



"""Unit tests for prompt, prompt_short, and LPS1 override behavior."""

import os
import unittest
from getpass import getuser
from unittest.mock import patch

from lshell.checkconfig import CheckConfig
from lshell.utils import getpromptbase, updateprompt

TOPDIR = f"{os.path.dirname(os.path.realpath(__file__))}/../"
CONFIG = f"{TOPDIR}/test/testfiles/test.conf"


class TestPromptUnit(unittest.TestCase):
    """Prompt rendering tests focused on documented behavior."""

    args = [f"--config={CONFIG}", "--quiet=1"]

    def test_prompt_short_rejects_values_outside_documented_range(self):
        """Reject prompt_short values that are not 0, 1, or 2."""
        with self.assertRaises(SystemExit) as exc:
            CheckConfig(self.args + ["--prompt_short=3"]).returnconf()
        self.assertEqual(exc.exception.code, 1)

    def test_lps1_override_ignores_prompt_and_prompt_short(self):
        """LPS1 should fully override both prompt template and path style."""
        conf = CheckConfig(
            self.args + ["--prompt='%u@%h'", "--prompt_short=2"]
        ).returnconf()
        with patch.dict(os.environ, {"LPS1": "PROMPT> "}, clear=False):
            rendered = updateprompt("/tmp/lshell-path-that-should-not-appear", conf)
        self.assertEqual(rendered, "PROMPT> ")

    def test_getpromptbase_uses_config_prompt_when_lps1_not_set(self):
        """Prompt placeholders should expand from config when LPS1 is absent."""
        conf = CheckConfig(self.args + ["--prompt='%u@%h'"]).returnconf()
        with patch.dict(os.environ, {}, clear=True):
            rendered = getpromptbase(conf)
        expected = f"{getuser()}@{os.uname()[1].split('.')[0]}"
        self.assertEqual(rendered, expected)


if __name__ == "__main__":
    unittest.main()



"""Functional attacker/sysadmin behavior tests for lshell sessions."""

import os
import unittest
from getpass import getuser

import pexpect


TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
CONFIG = f"{TOPDIR}/test/testfiles/test.conf"
LSHELL = f"{TOPDIR}/bin/lshell"
USER = getuser()
PROMPT = f"{USER}:~\\$"


class TestUserBehaviorSecurityFunctional(unittest.TestCase):
    """End-to-end tests that mimic realistic operator and attacker behavior."""

    def _spawn_shell(self, extra_args=""):
        """Spawn lshell and wait for prompt."""
        child = pexpect.spawn(
            f"{LSHELL} --config {CONFIG} {extra_args}",
            encoding="utf-8",
            timeout=10,
        )
        child.expect(PROMPT)
        return child

    def _run_command(self, child, command):
        """Run one command and return command output block."""
        child.sendline(command)
        child.expect(PROMPT)
        return child.before.split("\n", 1)[1].strip()

    def _contains_standalone_line(self, text, expected_line):
        """Return True when expected_line appears as its own output line."""
        return expected_line in [line.strip() for line in text.splitlines()]

    def _exit_shell(self, child):
        """Exit shell session and wait for EOF."""
        if not child.isalive():
            return
        child.sendline("exit")
        child.expect(pexpect.EOF)

    def test_hacker_session_exhausts_warning_budget_and_is_kicked(self):
        """Repeated policy violations should consume warnings and terminate session."""
        child = self._spawn_shell("--strict 1 --warning_counter 2 --quiet 0")
        try:
            first_probe = self._run_command(child, "echo SAFE; echo PWN")
            self.assertIn('lshell: forbidden character: ";"', first_probe)
            self.assertIn("lshell: warning: 1 violation", first_probe)

            second_probe = self._run_command(child, "id")
            self.assertIn('lshell: forbidden command: "id"', second_probe)
            self.assertIn("lshell: warning: 0 violations", second_probe)

            child.sendline("id")
            child.expect(pexpect.EOF)
            child.close()
            self.assertEqual(child.exitstatus, 1)
        finally:
            child.close()

    def test_inline_path_hijack_attempt_is_blocked_and_session_continues(self):
        """PATH=... command prefix should be rejected as forbidden env manipulation."""
        child = self._spawn_shell("--strict 0 --forbidden \"[]\" --allowed \"+['id']\"")
        try:
            hijack_attempt = self._run_command(child, "PATH=/tmp id")
            self.assertIn("lshell: forbidden environment variable: PATH", hijack_attempt)

            still_usable = self._run_command(child, "echo still_here")
            self.assertIn("still_here", still_usable)
            self._exit_shell(child)
        finally:
            child.close()

    def test_operator_smuggling_is_rejected_without_running_payload(self):
        """Malformed operator chains should fail closed and skip payload execution."""
        child = self._spawn_shell("--strict 0 --forbidden \"[]\" --allowed \"+['printf']\"")
        try:
            smuggling_attempt = self._run_command(child, "printf SAFE ||| printf PWNED")
            self.assertIn("lshell: unknown syntax:", smuggling_attempt)
            self.assertFalse(
                self._contains_standalone_line(smuggling_attempt, "PWNED"),
                msg="payload output should not execute for malformed operator chains",
            )

            post_attack = self._run_command(child, "echo AFTER")
            self.assertIn("AFTER", post_attack)
            self._exit_shell(child)
        finally:
            child.close()



"""Unit tests for parser utilities and job-control built-ins."""

import io
import os
import tempfile
import unittest
from contextlib import redirect_stdout
from unittest.mock import patch

from lshell import builtincmd
from lshell import completion
from lshell import utils


class FakeJob:
    """Simple fake subprocess object for background job tests."""

    def __init__(self, poll_value=None, returncode=0, pid=12345, args=None, cmd=None):
        self._poll_value = poll_value
        self.returncode = returncode
        self.pid = pid
        self.args = args or ["sleep", "60"]
        self.lshell_cmd = cmd or "sleep 60"
        self.wait_called = False

    def poll(self):
        """Return the configured poll status."""
        return self._poll_value

    def wait(self):
        """Mark the fake job as waited and return its final code."""
        self.wait_called = True
        self._poll_value = self.returncode
        return self.returncode


class FakeProcess:
    """Simple fake subprocess object used by exec_cmd signal tests."""

    def __init__(self, pid=12345, returncode=0, trigger_suspend=False):
        self.pid = pid
        self.returncode = returncode
        self.trigger_suspend = trigger_suspend
        self.lshell_cmd = ""
        self.args = ["tail", "-f", "blabla"]
        self._poll_value = None
        self._signal_handlers = {}

    def poll(self):
        """Report running state until the fake process completes."""
        return self._poll_value

    def communicate(self):
        """Simulate command execution, optionally triggering Ctrl+Z."""
        if self.trigger_suspend:
            handler = self._signal_handlers.get(utils.signal.SIGTSTP)
            if handler is not None:
                handler(utils.signal.SIGTSTP, None)
        self._poll_value = self.returncode


class TestParserUtilities(unittest.TestCase):
    """Tests for shell command parsing and env expansion helpers."""

    def test_split_command_sequence_keeps_quoted_operators(self):
        """Keep logical operators literal when they are inside quotes."""
        line = 'echo "a&&b||c" && echo done'
        self.assertEqual(utils.split_command_sequence(line), ['echo "a&&b||c"', "&&", "echo done"])

    def test_split_command_sequence_keeps_substitution_content(self):
        """Avoid splitting on operators located inside command substitution."""
        line = "echo $(printf '%s' 'a|b') | wc -c"
        self.assertEqual(
            utils.split_command_sequence(line),
            ["echo $(printf '%s' 'a|b')", "|", "wc -c"],
        )

    def test_split_command_sequence_keeps_parameter_expansion_with_logical_tokens(self):
        """Do not split on && / || that appear inside ${...} expansions."""
        line = "echo ${LSHELL_WORD:-a||b&&c} && echo done"
        self.assertEqual(
            utils.split_command_sequence(line),
            ["echo ${LSHELL_WORD:-a||b&&c}", "&&", "echo done"],
        )

    def test_split_command_sequence_keeps_nested_command_substitution(self):
        """Nested $(...) blocks should be treated as one command operand."""
        line = "echo $(printf '%s' \"$(echo a|tr a b)\") | wc -c"
        self.assertEqual(
            utils.split_command_sequence(line),
            ["echo $(printf '%s' \"$(echo a|tr a b)\")", "|", "wc -c"],
        )

    def test_split_command_sequence_rejects_unbalanced_quote(self):
        """Reject command lines with unmatched quotes."""
        self.assertIsNone(utils.split_command_sequence('echo "unterminated'))

    def test_split_command_sequence_does_not_split_redirection_ampersand(self):
        """Keep redirection forms like 2>&1 in the same command token."""
        line = "echo hi >/tmp/out 2>&1"
        self.assertEqual(utils.split_command_sequence(line), [line])

    @patch.dict(os.environ, {"A": "VALUE_A", "B": "VALUE_B"}, clear=False)
    def test_expand_vars_quoted_respects_single_quotes(self):
        """Expand variables except those protected by single quotes."""
        line = "echo $A '$A' \"$B\" ${A}"
        self.assertEqual(
            utils.expand_vars_quoted(line),
            "echo VALUE_A '$A' \"VALUE_B\" VALUE_A",
        )

    @patch.dict(os.environ, {"A": "VALUE_A"}, clear=False)
    def test_expand_vars_quoted_keeps_escaped_dollar_literal(self):
        """Backslash-escaped dollars should remain literal."""
        line = r"echo \$A \"$A\""
        self.assertEqual(utils.expand_vars_quoted(line), r"echo \$A \"VALUE_A\"")

    @patch.dict(os.environ, {"LSHELL_SET": "set"}, clear=False)
    def test_expand_vars_quoted_should_support_shell_parameter_default_expansion(self):
        """Expected shell parity: expand default value forms in ${...} expressions."""
        self.assertEqual(
            utils.expand_vars_quoted("echo ${LSHELL_UNSET:-fallback}"),
            "echo fallback",
        )
        self.assertEqual(
            utils.expand_vars_quoted("echo ${LSHELL_SET:-fallback}"),
            "echo set",
        )

    @patch.dict(os.environ, {"LSHELL_SET": "set"}, clear=False)
    def test_expand_vars_quoted_should_support_shell_parameter_alternative_expansion(self):
        """Expected shell parity: support ${VAR:+alt} semantics."""
        self.assertEqual(
            utils.expand_vars_quoted("echo ${LSHELL_SET:+ALT}"),
            "echo ALT",
        )
        self.assertEqual(
            utils.expand_vars_quoted("echo ${LSHELL_UNSET:+ALT}"),
            "echo ",
        )

    @patch.dict(os.environ, {"LSHELL_LEN": "abcd"}, clear=False)
    def test_expand_vars_quoted_should_support_shell_parameter_length_expansion(self):
        """Expected shell parity: support ${#VAR} length expansion."""
        self.assertEqual(
            utils.expand_vars_quoted("echo ${#LSHELL_LEN}"),
            "echo 4",
        )

    def test_parse_command_extracts_assignments_and_command(self):
        """Split assignments from executable and arguments."""
        executable, argument, split, assignments = utils._parse_command(
            "A=1 B=two echo hello world"
        )
        self.assertEqual(executable, "echo")
        self.assertEqual(argument, "hello world")
        self.assertEqual(split, ["A=1", "B=two", "echo", "hello", "world"])
        self.assertEqual(assignments, [("A", "1"), ("B", "two")])

    def test_parse_command_assignment_only(self):
        """Handle assignment-only input with no executable command."""
        executable, argument, split, assignments = utils._parse_command("A=1 B=two")
        self.assertEqual(executable, "")
        self.assertEqual(argument, "")
        self.assertEqual(split, ["A=1", "B=two"])
        self.assertEqual(assignments, [("A", "1"), ("B", "two")])

    def test_complete_list_dir_filters_by_prefix(self):
        """List completion should only return entries matching the typed prefix."""
        with tempfile.TemporaryDirectory() as tmpdir:
            os.mkdir(os.path.join(tmpdir, "blipdir"))
            open(os.path.join(tmpdir, "blabla"), "w", encoding="utf-8").close()
            open(os.path.join(tmpdir, "alpha"), "w", encoding="utf-8").close()
            conf = {"home_path": tmpdir, "path": ["", ""]}
            with patch("lshell.completion.os.getcwd", return_value=tmpdir), patch(
                "lshell.completion.sec.check_path", return_value=(0, conf)
            ):
                result = completion.complete_list_dir(
                    conf, "bl", "tail -f bl", 8, 10
                )

        self.assertIn("blipdir/", result)
        self.assertIn("blabla ", result)
        self.assertNotIn("alpha ", result)

    def test_complete_list_dir_filters_for_prefixed_path(self):
        """Completion should filter basenames when the token includes a directory path."""
        with tempfile.TemporaryDirectory() as tmpdir:
            os.mkdir(os.path.join(tmpdir, "blipdir"))
            open(os.path.join(tmpdir, "blabla"), "w", encoding="utf-8").close()
            open(os.path.join(tmpdir, "beta"), "w", encoding="utf-8").close()
            token = os.path.join(tmpdir, "bl")
            line = f"tail -f {token}"
            begidx = line.rfind(token)
            endidx = len(line)
            conf = {"home_path": tmpdir, "path": ["", ""]}
            with patch("lshell.completion.sec.check_path", return_value=(0, conf)):
                result = completion.complete_list_dir(
                    conf, token, line, begidx, endidx
                )

        self.assertIn("blipdir/", result)
        self.assertIn("blabla ", result)
        self.assertNotIn("beta ", result)

    def test_complete_list_dir_empty_token_lists_current_directory(self):
        """Trailing-space completion should list all entries in the working directory."""
        with tempfile.TemporaryDirectory() as tmpdir:
            os.mkdir(os.path.join(tmpdir, "lshell"))
            open(os.path.join(tmpdir, "bla"), "w", encoding="utf-8").close()
            open(os.path.join(tmpdir, "blabla"), "w", encoding="utf-8").close()
            open(os.path.join(tmpdir, "123"), "w", encoding="utf-8").close()
            conf = {"home_path": tmpdir, "path": ["", ""]}
            with patch("lshell.completion.os.getcwd", return_value=tmpdir), patch(
                "lshell.completion.sec.check_path", return_value=(0, conf)
            ):
                result = completion.complete_list_dir(
                    conf, "", "tail -f ", 8, 8
                )

        self.assertIn("lshell/", result)
        self.assertIn("bla ", result)
        self.assertIn("blabla ", result)
        self.assertIn("123 ", result)

    def test_complete_list_dir_subdirectory_with_empty_text(self):
        """Slash-delimited completion should list entries from the selected subdirectory."""
        with tempfile.TemporaryDirectory() as tmpdir:
            subdir = os.path.join(tmpdir, "lshell")
            os.mkdir(subdir)
            open(os.path.join(subdir, "README.md"), "w", encoding="utf-8").close()
            os.mkdir(os.path.join(subdir, "test"))
            open(os.path.join(tmpdir, ".bashrc"), "w", encoding="utf-8").close()
            conf = {"home_path": tmpdir, "path": ["", ""]}
            line = "tail -f lshell/"
            begidx = len(line)
            endidx = len(line)
            with patch("lshell.completion.os.getcwd", return_value=tmpdir), patch(
                "lshell.completion.sec.check_path", return_value=(0, conf)
            ):
                result = completion.complete_list_dir(conf, "", line, begidx, endidx)

        self.assertIn("README.md ", result)
        self.assertIn("test/", result)
        self.assertNotIn(".bashrc ", result)

    def test_complete_list_dir_subdirectory_prefix_completion(self):
        """Path-prefix completion should resolve candidates from the subdirectory."""
        with tempfile.TemporaryDirectory() as tmpdir:
            subdir = os.path.join(tmpdir, "lshell")
            os.mkdir(subdir)
            os.mkdir(os.path.join(subdir, "rpm"))
            os.mkdir(os.path.join(subdir, "source"))
            open(os.path.join(subdir, "README.md"), "w", encoding="utf-8").close()
            conf = {"home_path": tmpdir, "path": ["", ""]}
            line = "tail -f lshell/rp"
            begidx = len("tail -f lshell/")
            endidx = len(line)
            with patch("lshell.completion.os.getcwd", return_value=tmpdir), patch(
                "lshell.completion.sec.check_path", return_value=(0, conf)
            ):
                result = completion.complete_list_dir(conf, "rp", line, begidx, endidx)

        self.assertEqual(result, ["rpm/"])

    def test_complete_list_dir_returns_empty_when_path_denied(self):
        """Denied paths should return an empty completion list."""
        conf = {"home_path": "/tmp", "path": ["", ""]}
        with patch(
            "lshell.completion.sec.check_path", return_value=(1, conf)
        ), patch("lshell.completion.os.getcwd", return_value="/tmp"):
            result = completion.complete_list_dir(conf, "bl", "tail -f bl", 8, 10)

        self.assertEqual(result, [])

    def test_complete_change_dir_denied_path_does_not_suggest_root_slash(self):
        """Denied cd completion should not suggest a standalone '/' segment."""
        conf = {
            "home_path": "/home/testuser",
            "path": ["/var/log/|", ""],
        }
        with patch("lshell.completion.os.getcwd", return_value="/home/testuser"), patch(
            "lshell.completion.sec.check_path", return_value=(1, conf)
        ):
            result = completion.complete_change_dir(
                conf, "", "cd /var/log/", len("cd "), len("cd /var/log/")
            )

        self.assertNotIn("/", result)

    def test_completenames_dot_slash_with_basename_text(self):
        """Complete ./ commands when readline provides text without ./ prefix."""
        conf = {"allowed": ["ls", "./shutdown.sh", "./show.sh"]}
        result = completion.completenames(conf, "shut", "./shut", 0, 6)
        self.assertEqual(result, ["shutdown.sh"])

    def test_completenames_dot_slash_with_prefixed_text(self):
        """Complete ./ commands when readline provides text including ./ prefix."""
        conf = {"allowed": ["ls", "./shutdown.sh", "./show.sh"]}
        result = completion.completenames(conf, "./shut", "./shut", 0, 6)
        self.assertEqual(result, ["./shutdown.sh"])


class TestBuiltinsJobsAndSource(unittest.TestCase):
    """Tests for built-in commands around job control."""

    def setUp(self):
        """Save and clear global background job state before each test."""
        self._previous_jobs = list(builtincmd.BACKGROUND_JOBS)
        builtincmd.BACKGROUND_JOBS.clear()

    def tearDown(self):
        """Restore global background job state after each test."""
        builtincmd.BACKGROUND_JOBS.clear()
        builtincmd.BACKGROUND_JOBS.extend(self._previous_jobs)

    def test_cmd_bg_fg_no_jobs(self):
        """Report failure when attempting fg with no jobs queued."""
        stdout = io.StringIO()
        with redirect_stdout(stdout):
            ret = builtincmd.cmd_bg_fg("fg", "")
        self.assertEqual(ret, 1)
        self.assertIn("no such job", stdout.getvalue())

    def test_cmd_bg_fg_invalid_job_id(self):
        """Reject non-numeric fg job identifiers."""
        builtincmd.BACKGROUND_JOBS.append(FakeJob())
        stdout = io.StringIO()
        with redirect_stdout(stdout):
            ret = builtincmd.cmd_bg_fg("fg", "abc")
        self.assertEqual(ret, 1)
        self.assertIn("lshell: invalid job ID", stdout.getvalue())

    @patch("os.getpgid", return_value=4242)
    @patch("os.killpg")
    def test_cmd_bg_fg_resumes_and_removes_job(self, mock_killpg, _mock_getpgid):
        """Resume a running job in foreground and remove it once completed."""
        job = FakeJob(poll_value=None, returncode=0, pid=9876, cmd="sleep 10")
        builtincmd.BACKGROUND_JOBS.append(job)
        stdout = io.StringIO()

        with redirect_stdout(stdout):
            ret = builtincmd.cmd_bg_fg("fg", "1")

        self.assertEqual(ret, 0)
        self.assertTrue(job.wait_called)
        self.assertEqual(len(builtincmd.BACKGROUND_JOBS), 0)
        self.assertIn("sleep 10", stdout.getvalue())
        mock_killpg.assert_called_once()

    @patch("lshell.builtincmd.os.getpgid", return_value=4242)
    @patch("lshell.builtincmd.os.killpg")
    def test_cmd_bg_fg_ctrl_z_keeps_single_job_entry(self, mock_killpg, _mock_getpgid):
        """Ctrl+Z during fg should not duplicate the same job or raise."""
        job = FakeJob(poll_value=None, returncode=0, pid=9876, cmd="tail -f blabla")
        builtincmd.BACKGROUND_JOBS.append(job)
        stdout = io.StringIO()
        handlers = {}

        def fake_signal(signum, handler):
            handlers[signum] = handler
            return None

        def trigger_ctrl_z():
            job.wait_called = True
            handlers[utils.signal.SIGTSTP](utils.signal.SIGTSTP, None)

        job.wait = trigger_ctrl_z

        with patch("lshell.builtincmd.signal.getsignal", return_value=None), patch(
            "lshell.builtincmd.signal.signal", side_effect=fake_signal
        ):
            with redirect_stdout(stdout):
                ret = builtincmd.cmd_bg_fg("fg", "1")

        self.assertEqual(ret, 0)
        self.assertTrue(job.wait_called)
        self.assertEqual(len(builtincmd.BACKGROUND_JOBS), 1)
        self.assertIn("Stopped", stdout.getvalue())
        self.assertEqual(mock_killpg.call_count, 2)

    @patch("lshell.utils.os.getpgid", return_value=4242)
    @patch("lshell.utils.os.killpg")
    def test_exec_cmd_ctrl_z_restores_handlers_and_deduplicates_job(
        self, _mock_killpg, _mock_getpgid
    ):
        """exec_cmd should restore original handlers and avoid duplicate job entries."""
        fake_proc = FakeProcess(pid=9876, trigger_suspend=True)
        builtincmd.BACKGROUND_JOBS.append(fake_proc)
        signal_handlers = {}
        initial_sigtstp = object()
        initial_sigcont = object()

        def fake_getsignal(signum):
            if signum == utils.signal.SIGTSTP:
                return initial_sigtstp
            if signum == utils.signal.SIGCONT:
                return initial_sigcont
            return None

        def fake_signal(signum, handler):
            signal_handlers[signum] = handler
            return None

        fake_proc._signal_handlers = signal_handlers
        with patch("lshell.utils.signal.getsignal", side_effect=fake_getsignal), patch(
            "lshell.utils.signal.signal", side_effect=fake_signal
        ), patch("lshell.utils.subprocess.Popen", return_value=fake_proc):
            ret = utils.exec_cmd("tail -f blabla")

        self.assertEqual(ret, 0)
        self.assertEqual(len(builtincmd.BACKGROUND_JOBS), 1)
        self.assertIs(signal_handlers[utils.signal.SIGTSTP], initial_sigtstp)
        self.assertIs(signal_handlers[utils.signal.SIGCONT], initial_sigcont)

    def test_cmd_jobs_displays_symbols(self):
        """Render job list with expected current and previous markers."""
        builtincmd.BACKGROUND_JOBS.extend(
            [
                FakeJob(poll_value=None, cmd="sleep 1"),
                FakeJob(poll_value=None, cmd="sleep 2"),
                FakeJob(poll_value=None, cmd="sleep 3"),
            ]
        )
        stdout = io.StringIO()
        with redirect_stdout(stdout):
            ret = builtincmd.cmd_jobs()

        self.assertEqual(ret, 0)
        output = stdout.getvalue()
        self.assertIn("[1]   Stopped        sleep 1", output)
        self.assertIn("[2]-  Stopped        sleep 2", output)
        self.assertIn("[3]+  Stopped        sleep 3", output)

    def test_check_background_jobs_removes_completed_job(self):
        """Drop completed jobs and print completion status."""
        builtincmd.BACKGROUND_JOBS.append(
            FakeJob(poll_value=0, returncode=0, cmd="sleep 1")
        )
        stdout = io.StringIO()
        with redirect_stdout(stdout):
            builtincmd.check_background_jobs()

        self.assertEqual(len(builtincmd.BACKGROUND_JOBS), 0)
        self.assertIn("Done", stdout.getvalue())

    def test_check_background_jobs_suppresses_user_interrupted_job_message(self):
        """Do not print completion output for user-interrupted jobs."""
        builtincmd.BACKGROUND_JOBS.append(
            FakeJob(poll_value=130, returncode=-2, cmd="sleep 1")
        )
        stdout = io.StringIO()
        with redirect_stdout(stdout):
            builtincmd.check_background_jobs()

        self.assertEqual(len(builtincmd.BACKGROUND_JOBS), 0)
        self.assertEqual(stdout.getvalue(), "")

    def test_check_background_jobs_prunes_all_completed_entries(self):
        """Completed jobs should all be removed without skipping entries."""
        builtincmd.BACKGROUND_JOBS.extend(
            [
                FakeJob(poll_value=0, returncode=0, cmd="sleep 1"),
                FakeJob(poll_value=1, returncode=1, cmd="sleep 2"),
                FakeJob(poll_value=0, returncode=0, cmd="sleep 3"),
            ]
        )
        stdout = io.StringIO()
        with redirect_stdout(stdout):
            builtincmd.check_background_jobs()

        self.assertEqual(len(builtincmd.BACKGROUND_JOBS), 0)
        output = stdout.getvalue()
        self.assertEqual(output.count("Done"), 2)
        self.assertEqual(output.count("Failed"), 1)

    def test_jobs_prunes_finished_and_reindexes_running(self):
        """jobs() should drop non-running jobs and reindex active ones from 1."""
        builtincmd.BACKGROUND_JOBS.extend(
            [
                FakeJob(poll_value=0, returncode=0, cmd="done job"),
                FakeJob(poll_value=1, returncode=1, cmd="failed job"),
                FakeJob(poll_value=None, returncode=0, cmd="running job"),
            ]
        )

        joblist = builtincmd.jobs()

        self.assertEqual(joblist, [[1, "Stopped", "running job"]])
        self.assertEqual(len(builtincmd.BACKGROUND_JOBS), 1)
        self.assertEqual(builtincmd.BACKGROUND_JOBS[0].lshell_cmd, "running job")
